from typing import Dict, List

import pandas as pd
import webviz_core_components as wcc
from dash import Input, Output, callback
from dash.development.base_component import Component
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from .._plugin_ids import PluginIds


class Filter(SettingsGroupABC):
    class Ids(str):
        ENSEMBLE = "ensemble"
        WELLS = "wells"
        MAX_NUMBER_OF_WELLS_SLIDER = "max-number-of-wells-slider"
        SORT_BY = "sort-by"
        ASCENDING_DESCENDING = "ascending-descending"

    def __init__(self, bhp_df: pd.DataFrame) -> None:
        super().__init__("Filter")

        self.bhp_df = bhp_df

    @property
    def ensembles(self) -> List[str]:
        return list(self.bhp_df["ENSEMBLE"].unique())

    @property
    def label_map(self) -> Dict[str, str]:
        return {
            "Mean": "mean",
            "Count (data points)": "count",
            "Stddev": "std",
            "Minimum": "min",
            "Maximum": "max",
            "P10 (high)": "high_p10",
            "P50": "p50",
            "P90 (low)": "low_p90",
        }

    @property
    def wells(self) -> List[set]:
        return sorted(
            list(set(col[5:] for col in self.bhp_df.columns if col.startswith("WBHP:")))
        )

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Ensemble",
                id=self.register_component_unique_id(Filter.Ids.ENSEMBLE),
                options=[{"label": i, "value": i} for i in self.ensembles],
                value=self.ensembles[0],
                clearable=False,
                multi=False,
            ),
            wcc.Dropdown(
                label="Sort by",
                id=self.register_component_unique_id(Filter.Ids.SORT_BY),
                options=[
                    {"label": key, "value": value}
                    for key, value in self.label_map.items()
                ],
                clearable=False,
                value="low_p90",
            ),
            wcc.RadioItems(
                vertical=False,
                id=self.register_component_unique_id(Filter.Ids.ASCENDING_DESCENDING),
                options=[
                    {"label": "Ascending", "value": True},
                    {"label": "Descending", "value": False},
                ],
                value=True,
                labelStyle={"display": "inline-block"},
            ),
            wcc.Slider(
                label="Max number of wells in plot",
                id=self.register_component_unique_id(
                    Filter.Ids.MAX_NUMBER_OF_WELLS_SLIDER
                ),
                min=1,
                max=len(self.wells),
                step=1,
                value=min(10, len(self.wells)),
                marks={1: 1, len(self.wells): len(self.wells)},
            ),
            wcc.SelectWithLabel(
                label="Wells",
                id=self.register_component_unique_id(Filter.Ids.WELLS),
                options=[{"label": i, "value": i} for i in self.wells],
                size=min([len(self.wells), 20]),
                value=self.wells,
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_ENSEMBLE), "data"
            ),
            Input(self.component_unique_id(Filter.Ids.ENSEMBLE).to_string(), "value"),
        )
        def _set_ensembles(selected_ensemble: str) -> str:
            return selected_ensemble

        @callback(
            Output(self.get_store_unique_id(PluginIds.Stores.SELECTED_SORT_BY), "data"),
            Input(self.component_unique_id(Filter.Ids.SORT_BY).to_string(), "value"),
        )
        def _set_sort_by(selected_sort_by: List[str]) -> List[str]:
            return selected_sort_by

        @callback(
            Output(
                self.get_store_unique_id(
                    PluginIds.Stores.SELECTED_ASCENDING_DESCENDING
                ),
                "data",
            ),
            Input(
                self.component_unique_id(Filter.Ids.ASCENDING_DESCENDING).to_string(),
                "value",
            ),
        )
        def _set_ascending_descending(selected_ascending_descending: str) -> str:
            return selected_ascending_descending

        @callback(
            Output(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_MAX_NUMBER_OF_WELLS),
                "data",
            ),
            Input(
                self.component_unique_id(
                    Filter.Ids.MAX_NUMBER_OF_WELLS_SLIDER
                ).to_string(),
                "value",
            ),
        )
        def _set_max_number_of_wells(
            max_number_of_wells_slider: List[int],
        ) -> List[int]:
            return max_number_of_wells_slider

        @callback(
            Output(self.get_store_unique_id(PluginIds.Stores.SELECTED_WELLS), "data"),
            Input(self.component_unique_id(Filter.Ids.WELLS).to_string(), "value"),
        )
        def _set_wells(selected_wells: List[str]) -> List[str]:
            if selected_wells == []:
                raise PreventUpdate
            return selected_wells


# ADD SELECT STATISTICS
class ViewSettings(SettingsGroupABC):
    class Ids(StrEnum):
        SELECT_STATISTICS = "select-statistics"

    def __init__(self) -> None:
        super().__init__("View Settings")

    @property
    def label_map(self) -> Dict[str, str]:
        return {
            "Mean": "mean",
            "Count (data points)": "count",
            "Stddev": "std",
            "Minimum": "min",
            "Maximum": "max",
            "P10 (high)": "high_p10",
            "P50": "p50",
            "P90 (low)": "low_p90",
        }

    def layout(self) -> List[Component]:
        return [
            wcc.SelectWithLabel(
                label="Select statistics",
                id=self.register_component_unique_id(
                    ViewSettings.Ids.SELECT_STATISTICS
                ),
                options=[
                    {"label": key, "value": value}
                    for key, value in self.label_map.items()
                ],
                size=8,
                value=["min", "low_p90", "mean", "high_p10", "max"],
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_STATISTICS), "data"
            ),
            Input(
                self.component_unique_id(
                    ViewSettings.Ids.SELECT_STATISTICS
                ).to_string(),
                "value",
            ),
        )
        def _set_select_statistics(selected_statistics: List[str]) -> List[str]:
            return selected_statistics



from enum import Enum, auto
from typing import Any, Callable, List, Optional

import pandas as pd
import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import dash_table, dcc, html

from ._business_logic import SwatinitQcDataModel
from ._markdown import (
    check_swatinit_description,
    pc_columns_description,
    qc_flag_description,
)


# pylint: disable = too-few-public-methods
class LayoutElements(str, Enum):
    PLOT_WRAPPER = auto()
    PLOT_SELECTOR = auto()
    TABLE_WRAPPER = auto()
    PLOT_EQLNUM_SELECTOR = auto()
    TABLE_EQLNUM_SELECTOR = auto()
    FILTERS_CONTINOUS = auto()
    FILTERS_CONTINOUS_MAX_PC = auto()
    FILTERS_DISCRETE = auto()
    COLOR_BY = auto()
    MAX_POINTS = auto()
    HIGHLIGHT_ABOVE = auto()
    GROUPBY_EQLNUM = auto()
    SELECTED_TAB = auto()
    MAP_FIGURE = auto()
    MAIN_FIGURE = auto()


class LayoutStyle:
    MAIN_HEIGHT = "87vh"
    HEADER = {
        "font-size": "15px",
        "color": "black",
        "text-transform": "uppercase",
        "border-color": "black",
    }
    TABLE_HEADER = {"fontWeight": "bold"}
    TABLE_STYLE = {"max-height": MAIN_HEIGHT, "overflowY": "auto"}
    TABLE_CELL_WIDTH = 95
    TABLE_CELL_HEIGHT = "10px"
    TABLE_HIGHLIGHT = {"backgroundColor": "rgb(230, 230, 230)", "fontWeight": "bold"}
    TABLE_CSS = [
        {
            "selector": ".dash-spreadsheet tr",
            "rule": f"height: {TABLE_CELL_HEIGHT};",
        },
    ]


class Tabs:
    QC_PLOTS = "Water Initializaion QC plots"
    MAX_PC_SCALING = "Capillary pressure scaling"
    OVERVIEW = "Overview and Information"


def plugin_main_layout(get_uuid: Callable, datamodel: SwatinitQcDataModel) -> wcc.Tabs:

    return wcc.Tabs(
        id=get_uuid(LayoutElements.SELECTED_TAB),
        value=Tabs.OVERVIEW,
        children=[
            TabLayout(
                tab_label=Tabs.OVERVIEW,
                selections_layout=None,
                main_layout=OverviewTabLayout(get_uuid, datamodel).main_layout,
            ),
            TabLayout(
                tab_label=Tabs.QC_PLOTS,
                selections_layout=TabQqPlotLayout(
                    get_uuid, datamodel
                ).selections_layout,
                main_layout=html.Div(id=get_uuid(LayoutElements.PLOT_WRAPPER)),
            ),
            TabLayout(
                tab_label=Tabs.MAX_PC_SCALING,
                selections_layout=TabMaxPcInfoLayout(
                    get_uuid, datamodel
                ).selections_layout,
                main_layout=html.Div(id=get_uuid(LayoutElements.TABLE_WRAPPER)),
            ),
        ],
    )


class TabLayout(wcc.Tab):
    def __init__(
        self, tab_label: str, selections_layout: Optional[list], main_layout: list
    ) -> None:
        flex_children = []
        if selections_layout is not None:
            flex_children.append(
                wcc.Frame(
                    style={
                        "flex": 1,
                        "height": LayoutStyle.MAIN_HEIGHT,
                        "overflowY": "auto",
                    },
                    children=selections_layout,
                )
            )
        flex_children.append(
            wcc.Frame(
                style={"flex": 5, "height": LayoutStyle.MAIN_HEIGHT},
                color="white",
                highlight=False,
                children=main_layout,
            )
        )
        super().__init__(
            label=tab_label,
            value=tab_label,
            children=wcc.FlexBox(children=flex_children),
        )


class FullScreen(wcc.WebvizPluginPlaceholder):
    def __init__(self, children: List[Any]) -> None:
        super().__init__(buttons=["expand"], children=children)


class TabQqPlotLayout:
    class MainPlots(str, Enum):
        WATERFALL = auto()
        PROP_VS_DEPTH = auto()

    def __init__(self, get_uuid: Callable, datamodel: SwatinitQcDataModel) -> None:
        self.datamodel = datamodel
        self.get_uuid = get_uuid

    def main_layout(
        self,
        main_figure: go.Figure,
        map_figure: go.Figure,
        qc_volumes: dict,
    ) -> wcc.FlexBox:
        return wcc.FlexBox(
            children=[
                wcc.FlexColumn(
                    flex=4,
                    children=wcc.Graph(
                        style={"height": "85vh"},
                        id=self.get_uuid(LayoutElements.MAIN_FIGURE),
                        figure=main_figure,
                    ),
                ),
                wcc.FlexColumn(
                    flex=1,
                    children=[
                        FullScreen(
                            wcc.Graph(
                                responsive=True,
                                style={"height": "100%", "min-height": "45vh"},
                                id=self.get_uuid(LayoutElements.MAP_FIGURE),
                                figure=map_figure,
                            )
                        ),
                        self.info_box(qc_volumes, height="35vh"),
                    ],
                ),
            ]
        )

    @property
    def selections_layout(self) -> html.Div:
        return html.Div(
            children=[
                html.Div(
                    wcc.Dropdown(
                        label="Select QC-visualization:",
                        id=self.get_uuid(LayoutElements.PLOT_SELECTOR),
                        options=[
                            {
                                "label": "Waterfall plot for water vol changes",
                                "value": self.MainPlots.WATERFALL,
                            },
                            {
                                "label": "Reservoir properties vs Depth",
                                "value": self.MainPlots.PROP_VS_DEPTH,
                            },
                        ],
                        value=self.MainPlots.PROP_VS_DEPTH,
                        clearable=False,
                    ),
                    style={"margin-bottom": "15px"},
                ),
                wcc.Selectors(
                    label="Selections",
                    children=[
                        wcc.SelectWithLabel(
                            label="EQLNUM",
                            id=self.get_uuid(LayoutElements.PLOT_EQLNUM_SELECTOR),
                            options=[
                                {"label": ens, "value": ens}
                                for ens in self.datamodel.eqlnums
                            ],
                            value=self.datamodel.eqlnums[:1],
                            size=min(8, len(self.datamodel.eqlnums)),
                            multi=True,
                        ),
                        wcc.Dropdown(
                            label="Color by",
                            id=self.get_uuid(LayoutElements.COLOR_BY),
                            options=[
                                {"label": ens, "value": ens}
                                for ens in self.datamodel.color_by_selectors
                            ],
                            value="QC_FLAG",
                            clearable=False,
                        ),
                        wcc.Label("Max number of points:"),
                        dcc.Input(
                            id=self.get_uuid(LayoutElements.MAX_POINTS),
                            type="number",
                            value=5000,
                        ),
                    ],
                ),
                wcc.Selectors(
                    label="Filters",
                    children=[
                        wcc.SelectWithLabel(
                            label="QC_FLAG",
                            id={
                                "id": self.get_uuid(LayoutElements.FILTERS_DISCRETE),
                                "col": "QC_FLAG",
                            },
                            options=[
                                {"label": ens, "value": ens}
                                for ens in self.datamodel.qc_flag
                            ],
                            value=self.datamodel.qc_flag,
                            size=min(8, len(self.datamodel.qc_flag)),
                        ),
                        wcc.SelectWithLabel(
                            label="SATNUM",
                            id={
                                "id": self.get_uuid(LayoutElements.FILTERS_DISCRETE),
                                "col": "SATNUM",
                            },
                            options=[
                                {"label": ens, "value": ens}
                                for ens in self.datamodel.satnums
                            ],
                            value=self.datamodel.satnums,
                            size=min(8, len(self.datamodel.satnums)),
                        ),
                        range_filters(
                            self.get_uuid(LayoutElements.FILTERS_CONTINOUS),
                            self.datamodel,
                        ),
                    ],
                ),
            ],
        )

    @staticmethod
    def info_box(qc_vols: dict, height: str) -> html.Div:
        return html.Div(
            [
                wcc.Header("Information about selection", style=LayoutStyle.HEADER),
                html.Div(
                    "EQLNUMS:", style={"font-weight": "bold", "font-size": "15px"}
                ),
                html.Div(", ".join([str(x) for x in qc_vols["EQLNUMS"]])),
                html.Div(
                    "SATNUMS:", style={"font-weight": "bold", "font-size": "15px"}
                ),
                html.Div(", ".join([str(x) for x in qc_vols["SATNUMS"]])),
                html.Div(
                    html.Span("Reservoir Volume Difference:"),
                    style={"font-weight": "bold", "margin-top": "10px"},
                ),
                html.Div(
                    children=[
                        html.Div(line)
                        for line in [
                            f"Water Volume Diff: {qc_vols['WVOL_DIFF']/(10**6):.2f} Mrm3",
                            f"Water Volume Diff (%): {qc_vols['WVOL_DIFF_PERCENT']:.2f}",
                            f"HC Volume Diff (%): {qc_vols['HCVOL_DIFF_PERCENT']:.2f}",
                        ]
                    ],
                ),
            ],
            style={"height": height, "padding": "10px"},
        )


class TabMaxPcInfoLayout:
    def __init__(self, get_uuid: Callable, datamodel: SwatinitQcDataModel) -> None:
        self.datamodel = datamodel
        self.get_uuid = get_uuid

    def main_layout(
        self, dframe: pd.DataFrame, selectors: list, map_figure: go.Figure
    ) -> html.Div:
        return html.Div(
            children=[
                wcc.Header(
                    "Maximum capillary pressure scaling", style=LayoutStyle.HEADER
                ),
                wcc.FlexBox(
                    style={"margin-top": "10px", "height": "40vh"},
                    children=[
                        wcc.FlexColumn(
                            dcc.Markdown(pc_columns_description()),
                            flex=7,
                            style={"margin-right": "40px"},
                        ),
                        wcc.FlexColumn(
                            FullScreen(
                                wcc.Graph(
                                    style={"height": "100%", "min-height": "35vh"},
                                    figure=map_figure,
                                )
                            ),
                            flex=3,
                        ),
                    ],
                ),
                self.create_max_pc_table(dframe, text_columns=selectors),
            ]
        )

    @property
    def selections_layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.Selectors(
                    label="Selections",
                    children=[
                        html.Div(
                            wcc.RadioItems(
                                label="Split table by:",
                                id=self.get_uuid(LayoutElements.GROUPBY_EQLNUM),
                                options=[
                                    {"label": "SATNUM", "value": "SATNUM"},
                                    {"label": "SATNUM and EQLNUM", "value": "both"},
                                ],
                                value="SATNUM",
                            ),
                            style={"margin-bottom": "10px"},
                        ),
                        self.scaling_threshold,
                    ],
                ),
                wcc.Selectors(
                    label="Filters",
                    children=[
                        wcc.SelectWithLabel(
                            label="EQLNUM",
                            id=self.get_uuid(LayoutElements.TABLE_EQLNUM_SELECTOR),
                            options=[
                                {"label": ens, "value": ens}
                                for ens in self.datamodel.eqlnums
                            ],
                            value=self.datamodel.eqlnums,
                            size=min(8, len(self.datamodel.eqlnums)),
                            multi=True,
                        ),
                        range_filters(
                            self.get_uuid(LayoutElements.FILTERS_CONTINOUS_MAX_PC),
                            self.datamodel,
                        ),
                    ],
                ),
            ]
        )

    def create_max_pc_table(
        self, dframe: pd.DataFrame, text_columns: list
    ) -> dash_table:
        return DashTable(
            data=dframe.to_dict("records"),
            columns=[
                {
                    "name": i,
                    "id": i,
                    "type": "numeric" if i not in text_columns else "text",
                    "format": {"specifier": ".4~r"} if i not in text_columns else {},
                }
                for i in dframe.columns
            ],
            height="48vh",
            sort_action="native",
            fixed_rows={"headers": True},
            style_cell={
                "minWidth": LayoutStyle.TABLE_CELL_WIDTH,
                "maxWidth": LayoutStyle.TABLE_CELL_WIDTH,
                "width": LayoutStyle.TABLE_CELL_WIDTH,
            },
            style_data_conditional=[
                {
                    "if": {
                        "filter_query": f"{{{self.datamodel.COLNAME_THRESHOLD}}} > 0",
                    },
                    **LayoutStyle.TABLE_HIGHLIGHT,
                },
            ],
        )

    @property
    def scaling_threshold(self) -> html.Div:
        return html.Div(
            style={"margin-top": "10px"},
            children=[
                wcc.Label("Maximum PC_SCALING threshold"),
                html.Div(
                    dcc.Input(
                        id=self.get_uuid(LayoutElements.HIGHLIGHT_ABOVE),
                        type="number",
                        persistence=True,
                        persistence_type="session",
                    )
                ),
            ],
        )


class OverviewTabLayout:
    def __init__(self, get_uuid: Callable, datamodel: SwatinitQcDataModel):
        self.datamodel = datamodel
        self.get_uuid = get_uuid

    @property
    def main_layout(self) -> html.Div:
        return html.Div(
            children=[
                html.Div(
                    style={"height": "40vh", "overflow-y": "auto"},
                    children=[
                        wcc.FlexBox(
                            children=[
                                wcc.FlexColumn(
                                    [
                                        self.overview_report_volume_changes,
                                    ],
                                    flex=7,
                                    style={"margin-right": "20px"},
                                ),
                                wcc.FlexColumn(self.dataset_info, flex=3),
                            ],
                        ),
                    ],
                ),
                html.Div(
                    style={"margin-top": "20px"},
                    children=[
                        wcc.Header("QC_FLAG descriptions", style=LayoutStyle.HEADER),
                        dcc.Markdown(qc_flag_description()),
                    ],
                ),
            ]
        )

    @property
    def information_dialog(self) -> html.Div:
        title = "Plugin and 'check_swatinit' information"
        return html.Div(
            style={"margin-bottom": "30px"},
            children=[
                html.Button(
                    title,
                    style={"width": "100%", "background-color": "white"},
                    id=self.get_uuid("info-button"),
                ),
                wcc.Dialog(
                    title=title,
                    id=self.get_uuid("info-dialog"),
                    max_width="md",
                    open=False,
                    children=dcc.Markdown(check_swatinit_description()),
                ),
            ],
        )

    @property
    def dataset_info(self) -> html.Div:
        max_pc, min_pc = self.datamodel.pc_scaling_min_max
        wvol, hcvol = self.datamodel.vol_diff_total

        number_style = {
            "font-weight": "bold",
            "font-size": "17px",
            "margin-left": "20px",
        }
        data = [
            ("HC Volume difference:", f"{hcvol:.2f} %"),
            ("Water Volume difference:", f"{wvol:.2f} %"),
            ("Maximum Capillary Pressure scaling:", f"{max_pc:.1f}"),
            ("Minimum Capillary Pressure scaling:", f"{min_pc:.3g}"),
        ]

        return wcc.Frame(
            style={"height": "90%"},
            children=[
                wcc.Header("Information", style=LayoutStyle.HEADER),
                self.information_dialog,
                wcc.Header("Key numbers", style=LayoutStyle.HEADER),
                html.Div(
                    [
                        html.Div([text, html.Span(num, style=number_style)])
                        for text, num in data
                    ]
                ),
            ],
        )

    @property
    def overview_report_volume_changes(self) -> html.Div:
        data, columns = self.datamodel.table_data_qc_vol_overview()
        label = (
            "Table showing volume changes from SWATINIT to SWAT at Reservoir conditions"
        )
        return html.Div(
            children=[
                html.Div(
                    html.Label(label, className="webviz-underlined-label"),
                    style={"margin-bottom": "20px"},
                ),
                DashTable(
                    data=data,
                    columns=columns,
                    style_data_conditional=[
                        {
                            "if": {"row_index": [0, len(data) - 1]},
                            **LayoutStyle.TABLE_HIGHLIGHT,
                        },
                    ],
                ),
            ],
        )


def range_filters(uuid: str, datamodel: SwatinitQcDataModel) -> html.Div:
    dframe = datamodel.dframe
    filters = []
    for col in datamodel.filters_continuous:
        min_val, max_val = dframe[col].min(), dframe[col].max()
        filters.append(
            wcc.RangeSlider(
                label="Depth range" if col == "Z" else col,
                id={"id": uuid, "col": col},
                min=min_val,
                max=max_val,
                value=[min_val, max_val],
                marks={str(val): {"label": f"{val:.2f}"} for val in [min_val, max_val]},
                tooltip={"always_visible": False},
            )
        )
    return html.Div(filters)


class DashTable(dash_table.DataTable):
    def __init__(
        self, data: List[dict], columns: List[dict], height: str = "none", **kwargs: Any
    ) -> None:
        super().__init__(
            data=data,
            columns=columns,
            style_table={"height": height, **LayoutStyle.TABLE_STYLE},
            style_as_list_view=True,
            css=LayoutStyle.TABLE_CSS,
            style_header=LayoutStyle.TABLE_HEADER,
            **kwargs,
        )



from typing import Any, Dict, List

import pandas as pd
import plotly.graph_objs as go

from .._utils.colors import hex_to_rgba_str
from .px_figure import create_figure


class ScatterPlot:
    """Class to create a general scatterplot

    Input:
    * df: dataframe with values
    * response: y-axis parameter
    * param: x-axis parameter
    * color: marker color
    * title
    * plot_trendline (default False)
    """

    def __init__(
        self,
        df: pd.DataFrame,
        response: str,
        param: str,
        color: str,
        title: str,
        plot_trendline: bool = False,
    ):
        self._figure = (
            create_figure(
                plot_type="scatter",
                data_frame=df[["REAL", response, param]],
                x=param,
                y=response,
                trendline="ols"
                if plot_trendline and df[response].nunique() > 1
                else None,
                trendline_color_override="#243746",
                hover_data=["REAL", response, param],
                framed=False,
            )
            .update_layout(
                margin={"r": 20, "l": 20, "t": 60, "b": 20},
                title={"text": title, "x": 0.5},
                xaxis_title=None,
                yaxis_title=None,
            )
            .update_traces(marker_color=color)
        )

    @property
    def figure(self) -> Dict[str, Any]:
        return self._figure

    def update_color(self, color: str, opacity: float) -> None:
        """Update color for scatter plot"""
        for trace in self._figure["data"]:
            if trace["mode"] == "markers":
                trace["marker"].update(color=hex_to_rgba_str(color, opacity))
                trace["marker"]["line"].update(color=hex_to_rgba_str(color, 1))

    def add_trace(
        self,
        x_values: List[float],
        y_values: List[float],
        mode: str,
        color: str,
        text: str,
        dash: str = "solid",
        showlegend: bool = False,
    ) -> None:
        self._figure.add_trace(
            go.Scatter(
                x=x_values,
                y=y_values,
                mode=mode,
                line=dict(dash=dash, color=color),
                showlegend=showlegend,
                hovertext=text,
                name=text,
            )
        )

    def add_vertical_line_with_error(
        self, value: float, error: float, xmin: float, xmax: float
    ) -> None:
        self.add_trace(
            [xmin, xmax],
            [value, value],
            "lines",
            "#243746",
            "Observation",
            showlegend=True,
        )
        self.add_trace(
            [xmin, xmax],
            [value - error, value - error],
            "lines",
            "#243746",
            "Observation Error",
            dash="dash",
        )
        self.add_trace(
            [xmin, xmax],
            [value + error, value + error],
            "lines",
            "#243746",
            "Observation Error",
            dash="dash",
        )
        self._figure.update_layout(legend=dict(orientation="h"))



import json
from pathlib import Path
from typing import Callable, Iterable, List, Tuple, Union
from uuid import uuid4

import numpy as np
import pandas as pd
import webviz_core_components as wcc
from dash import Dash, Input, Output, State, callback_context, dash_table, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.deprecation_decorators import deprecated_plugin
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._components import TornadoWidget

from .._abbreviations.number_formatting import table_statistics_base
from .._abbreviations.volume_terminology import volume_description, volume_unit
from .._datainput.fmu_input import find_sens_type, get_realizations
from .._datainput.inplace_volumes import extract_volumes


@deprecated_plugin(
    "Relevant functionality is implemented in the VolumetricAnalysis plugin."
)
class InplaceVolumesOneByOne(WebvizPluginABC):
    """Visualizes inplace volumetrics related to a FMU ensemble with a design matrix.

Input can be given either as an aggregated `csv` file for volumes and sensitivity information,
or as ensemble name(s) defined in `shared_settings` and volumetric `csv` files
stored per realization.

---

* **`csvfile_vol`:** Aggregated csvfile for volumes with `REAL`, `ENSEMBLE` and `SOURCE` columns.
* **`csvfile_parameters`:** Aggregated csvfile of parameters for sensitivity information with \
  `REAL`, `ENSEMBLE`, `SENSNAME` and `SENSCASE` columns.
* **`ensembles`:** Which ensembles in `shared_settings` to visualize (not to be used with \
  `csvfile_vol` and `csvfile_parameters`).
* **`volfiles`:**  Key/value pair of csv files when using `ensembles`. \
  E.g. `{geogrid: geogrid--oil.csv}`.
* **`volfolder`:** Optional local folder for the `volfiles`.
* **`response`:** Optional volume response to visualize initially.

---
?> The input files must follow FMU standards.


**Volumetric input**

* [Example of an aggregated file for `csvfile_vol`](https://github.com/equinor/\
webviz-subsurface-testdata/blob/master/aggregated_data/volumes.csv).

* [Example of a file per realization that can be used with `ensembles` and `volfiles`]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_history_match/\
realization-0/iter-0/share/results/volumes/geogrid--oil.csv).

The following columns will be used as available filters, if present:

* `ZONE`
* `REGION`
* `FACIES`
* `LICENSE`
* `SOURCE` (relevant if calculations are done for multiple grids)


Remaining columns are seen as volumetric responses.

All names are allowed (except those mentioned above, in addition to `REAL` and `ENSEMBLE`), \
but the following responses are given more descriptive names automatically:

* `BULK_OIL`: Bulk Volume (Oil)
* `NET_OIL`: Net Volume (Oil)
* `PORE_OIL`: Pore Volume (Oil)
* `HCPV_OIL`: Hydro Carbon Pore Volume (Oil)
* `STOIIP_OIL`: Stock Tank Oil Initially In Place
* `BULK_GAS`: Bulk Volume (Gas)
* `NET_GAS`: Net Volume (Gas)
* `PORV_GAS`: Pore Volume (Gas)
* `HCPV_GAS`: Hydro Carbon Pore Volume (Gas)
* `GIIP_GAS`: Gas Initially In Place
* `RECOVERABLE_OIL`: Recoverable Volume (Oil)
* `RECOVERABLE_GAS`: Recoverable Volume (Gas)

**Sensitivity input**

The sensitivity information is extracted automatically if `ensembles` is given as input,
as long as `SENSCASE` and `SENSNAME` is found in `parameters.txt`.

An example of an aggregated file to use with `csvfile_parameters`
[can be found here](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
aggregated_data/parameters.csv)
"""

    FILTERS = ["ZONE", "REGION", "FACIES", "LICENSE"]

    TABLE_STATISTICS: List[Tuple[str, dict]] = [
        ("Sens Name", {}),
        ("Sens Case", {}),
    ] + table_statistics_base()

    ENSEMBLE_COLUMNS = [
        "REAL",
        "ENSEMBLE",
        "SOURCE",
        "SENSCASE",
        "SENSNAME",
        "SENSTYPE",
        "RUNPATH",
    ]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csvfile_vol: Path = None,
        csvfile_parameters: Path = None,
        ensembles: list = None,
        volfiles: dict = None,
        volfolder: str = "share/results/volumes",
        response: str = "STOIIP_OIL",
    ):

        super().__init__()

        self.csvfile_vol = csvfile_vol if csvfile_vol else None
        self.csvfile_parameters = csvfile_parameters if csvfile_parameters else None

        if csvfile_vol and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_vol" and "csvfile_parameters" or '
                '"ensembles" and "volfiles"'
            )
        if csvfile_vol and csvfile_parameters:
            volumes = read_csv(csvfile_vol)
            parameters = read_csv(csvfile_parameters)
            parameters["SENSTYPE"] = parameters.apply(
                lambda row: find_sens_type(row.SENSCASE), axis=1
            )

        elif ensembles and volfiles:
            self.ens_paths = {
                ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.volfiles = volfiles
            self.volfolder = volfolder
            # Extract volumetric dataframe
            volumes = extract_volumes(self.ens_paths, self.volfolder, self.volfiles)
            # Extract realizations and sensitivity information
            parameters = get_realizations(
                ensemble_paths=self.ens_paths, ensemble_set_name="EnsembleSet"
            )

        else:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_vol" and "csvfile_parameters" or '
                '"ensembles" and "volfiles"'
            )
        self.initial_response = response

        # Merge into one dataframe
        # (TODO: Should raise error if not all ensembles have sensitivity data)
        self.volumes = pd.merge(volumes, parameters, on=["ENSEMBLE", "REAL"])

        # Initialize a tornado plot. Data is added in callback
        self.tornadoplot = TornadoWidget(
            app, webviz_settings, parameters, allow_click=True
        )
        self.uid = uuid4()
        self.selectors_id = {x: self.uuid(x) for x in self.selectors}
        self.theme = webviz_settings.theme
        self.set_callbacks(app)

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return (
            [
                (
                    read_csv,
                    [
                        {"csv_file": self.csvfile_vol},
                        {"csv_file": self.csvfile_parameters},
                    ],
                )
            ]
            if self.csvfile_vol and self.csvfile_parameters
            else [
                (
                    extract_volumes,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "volfolder": self.volfolder,
                            "volfiles": self.volfiles,
                        }
                    ],
                ),
                (
                    get_realizations,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "ensemble_set_name": "EnsembleSet",
                        }
                    ],
                ),
            ]
        )

    def selector(self, label: str, id_name: str, column: str) -> html.Div:
        return wcc.Dropdown(
            label=label,
            id=self.uuid(id_name),
            options=[
                {"label": i, "value": i} for i in list(self.volumes[column].unique())
            ],
            clearable=False,
            value=list(self.volumes[column])[0],
        )

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": (
                    "Dashboard displaying in place volumetric results "
                    "from a sensitivity study."
                ),
            },
            {
                "id": self.uuid("graph-wrapper"),
                "content": (
                    "Chart showing results for the current selection. "
                    "Different charts and options can be selected from the menu above. "
                    "Different sensitivities can be highlighted by clicking in the tornado plot."
                ),
            },
            {
                "id": self.uuid("table"),
                "content": (
                    "The table shows statistics per sensitivity parameter. "
                    "Rows can be filtered by searching, and sorted by "
                    "clicking on a column header."
                ),
            },
            {
                "id": self.uuid("response"),
                "content": "Select the volumetric calculation to display.",
            },
            {
                "id": self.uuid("plot-type"),
                "content": (
                    "Controls the type of the visualized chart. "
                    "Per realization shows bars per realization, "
                    "while the boxplot shows the range per sensitivity."
                ),
            },
            *self.tornadoplot.tour_steps,
            {
                "id": self.uuid("ensemble"),
                "content": (
                    "If several ensembles are available, the active ensemble "
                    "can be selected here."
                ),
            },
            {
                "id": self.uuid("source"),
                "content": (
                    "If volumes have been calculated for different grids. "
                    "E.g. geogrid and eclipsegrid, the active grid can be selected here."
                ),
            },
            {
                "id": self.uuid("filters"),
                "content": (
                    "Filter on different combinations of e.g. zones, facies and regions "
                    "(The options will vary dependent on what was included "
                    "in the calculation.)"
                ),
            },
        ]

    @property
    def vol_columns(self) -> List[str]:
        """List of all columns in dataframe"""
        return list(self.volumes.columns)

    @property
    def selectors(self) -> List[str]:
        """List of available selector columns in dframe"""
        return [x for x in InplaceVolumesOneByOne.FILTERS if x in self.vol_columns]

    @property
    def responses(self) -> List[str]:
        """List of available volume responses in dframe"""
        return [
            x
            for x in self.vol_columns
            if x not in self.selectors
            and x not in InplaceVolumesOneByOne.ENSEMBLE_COLUMNS
        ]

    @property
    def plot_selector(self) -> html.Div:
        """Radiobuttons to select plot type"""
        return wcc.Dropdown(
            label="Plot type",
            id=self.uuid("plot-type"),
            options=[
                {"label": i, "value": i}
                for i in [
                    "Per realization",
                    "Per sensitivity name",
                    "Per sensitivity case",
                ]
            ],
            value="Per realization",
            clearable=False,
        )

    @property
    def response_selector(self) -> html.Div:
        """Dropdown to select volumetric response"""
        return wcc.Dropdown(
            label="Volumetric calculation",
            id=self.uuid("response"),
            options=[
                {"label": volume_description(i), "value": i} for i in self.responses
            ],
            clearable=False,
            value=self.initial_response
            if self.initial_response in self.responses
            else self.responses[0],
        )

    @property
    def filter_selectors(self) -> List[html.Div]:
        """Dropdowns for dataframe columns that can be filtered on (Zone, Region, etc)"""
        return [
            wcc.SelectWithLabel(
                label=selector,
                id=self.selectors_id[selector],
                options=[
                    {"label": i, "value": i}
                    for i in list(self.volumes[selector].unique())
                ],
                value=list(self.volumes[selector].unique()),
                multi=True,
                size=min(20, len(self.volumes[selector].unique())),
            )
            for selector in self.selectors
        ]

    @property
    def layout(self) -> html.Div:
        """Main layout"""
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                wcc.FlexColumn(
                    flex=1,
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        children=[
                            wcc.Selectors(
                                label="Selectors",
                                children=[
                                    self.selector(
                                        "Ensemble",
                                        "ensemble",
                                        "ENSEMBLE",
                                    ),
                                    self.selector(
                                        "Grid source",
                                        "source",
                                        "SOURCE",
                                    ),
                                    self.response_selector,
                                    self.plot_selector,
                                ],
                            ),
                            wcc.Selectors(
                                label="Filters",
                                id=self.uuid("filters"),
                                children=self.filter_selectors,
                            ),
                        ],
                    ),
                ),
                wcc.FlexColumn(
                    flex=3,
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        color="white",
                        highlight=False,
                        children=[
                            html.Div(
                                style={"height": "600px"},
                                id=self.uuid("graph-wrapper"),
                            ),
                            html.Div(
                                children=[
                                    html.Div(
                                        id=self.uuid("volume_title"),
                                        style={"textAlign": "center"},
                                        children="",
                                    ),
                                    html.Div(
                                        style={"fontSize": "15px"},
                                        children=dash_table.DataTable(
                                            id=self.uuid("table"),
                                            sort_action="native",
                                            filter_action="native",
                                            page_action="native",
                                            page_size=10,
                                        ),
                                    ),
                                ],
                            ),
                        ],
                    ),
                ),
                wcc.FlexColumn(
                    flex=3,
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        id=self.uuid("tornado-wrapper"),
                        color="white",
                        highlight=False,
                        children=[self.tornadoplot.layout],
                    ),
                ),
            ],
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            [
                Output(self.tornadoplot.storage_id, "data"),
                Output(self.uuid("table"), "data"),
                Output(self.uuid("table"), "columns"),
            ],
            [
                Input(i, "value")
                for sublist in [
                    [
                        self.uuid("ensemble"),
                        self.uuid("response"),
                        self.uuid("source"),
                    ],
                    list(self.selectors_id.values()),
                ]
                for i in sublist
            ],
        )
        def _render_table_and_tornado(
            ensemble: str, response: str, source: str, *filters: Union[str, List[str]]
        ) -> Tuple[str, List[dict], List[dict]]:
            # Filter data
            data = filter_dataframe(
                self.volumes, self.selectors, ensemble, source, filters
            )

            # Table data
            table, columns = calculate_table(data, response)

            # TornadoPlot input
            tornado = json.dumps(
                {
                    "ENSEMBLE": ensemble,
                    "data": data.groupby("REAL")
                    .sum()
                    .reset_index()[["REAL", response]]
                    .values.tolist(),
                    "response_name": response,
                    "number_format": "#.4g",
                    "unit": volume_unit(response),
                }
            )
            return tornado, table, columns

        @app.callback(
            [
                Output(self.uuid("graph-wrapper"), "children"),
                Output(self.uuid("volume_title"), "children"),
            ],
            [
                Input(self.tornadoplot.click_id, "data"),
                Input(self.tornadoplot.high_low_storage_id, "data"),
                Input(self.uuid("plot-type"), "value"),
            ],
            [
                State(i, "value")
                for sublist in [
                    [
                        self.uuid("ensemble"),
                        self.uuid("response"),
                        self.uuid("source"),
                    ],
                    list(self.selectors_id.values()),
                ]
                for i in sublist
            ],
        )
        def _render_chart(
            tornado_click_data_str: Union[str, None],
            high_low_storage: dict,
            plot_type: str,
            ensemble: str,
            response: str,
            source: str,
            *filters: Union[str, List[str]],
        ) -> Tuple[Union[wcc.Graph, html.Div], str]:
            if callback_context.triggered is None:
                raise PreventUpdate

            tornado_click: Union[dict, None] = (
                json.loads(tornado_click_data_str) if tornado_click_data_str else None
            )

            ctx = callback_context.triggered[0]["prop_id"].split(".")[0]
            if tornado_click:
                if (
                    high_low_storage is not None
                    and ctx == self.tornadoplot.high_low_storage_id
                ):
                    # Tornado plot is updated without a click in the plot, updating reals:
                    if tornado_click["sens_name"] in high_low_storage:
                        tornado_click["real_low"] = high_low_storage[
                            tornado_click["sens_name"]
                        ].get("real_low")
                        tornado_click["real_high"] = high_low_storage[
                            tornado_click["sens_name"]
                        ].get("real_high")
                    else:
                        # fallback in a case where chosen sens_name does not exist in updated
                        # tornado plot.
                        # (can this ever occur except when sens_name already is None?)
                        tornado_click["sens_name"] = None
                        tornado_click["real_low"] = []
                        tornado_click["real_high"] = []

            # Filter data
            data = filter_dataframe(
                self.volumes, self.selectors, ensemble, source, filters
            )

            # Volume title:
            volume_title = f"{volume_description(response)} [{volume_unit(response)}]"

            # Make Plotly figure
            layout = {}
            layout.update({"height": 600, "margin": {"l": 100, "b": 100}})
            if plot_type == "Per realization":
                # One bar per realization
                layout.update(
                    {
                        "title": "Response per realization",
                        "xaxis": {"title": "Realizations"},
                        "yaxis": {"title": volume_title},
                    }
                )
                plot_data = data.groupby("REAL").sum().reset_index()

                if tornado_click:
                    figure_data = [
                        {
                            "y": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_low"])
                            ][response],
                            "x": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_low"])
                            ]["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "showlegend": False,
                            "name": "",
                        },
                        {
                            "y": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_high"])
                            ][response],
                            "x": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_high"])
                            ]["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "showlegend": False,
                            "name": "",
                        },
                        {
                            "y": plot_data[
                                ~plot_data["REAL"].isin(
                                    tornado_click["real_low"]
                                    + tornado_click["real_high"]
                                )
                            ][response],
                            "x": plot_data[
                                ~plot_data["REAL"].isin(
                                    tornado_click["real_low"]
                                    + tornado_click["real_high"]
                                )
                            ]["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "marker": {"color": "grey"},
                            "showlegend": False,
                            "name": "",
                        },
                    ]
                else:
                    figure_data = [
                        {
                            "y": plot_data[response],
                            "x": plot_data["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "marker": {"color": "grey"},
                        },
                    ]
                figure = wcc.Graph(
                    config={"displayModeBar": False},
                    id=self.uuid("graph"),
                    figure={
                        "data": figure_data,
                        "layout": self.theme.create_themed_layout(layout),
                    },
                )
            elif plot_type == "Per sensitivity case":
                # One box per sensitivity name
                layout.update(
                    {
                        "title": "Distribution for each sensitivity case",
                        "yaxis": {"title": volume_title},
                    }
                )
                figure = wcc.Graph(
                    config={"displayModeBar": False},
                    id=self.uuid("graph"),
                    figure={
                        "data": [
                            {
                                "y": senscase_df.groupby("REAL")
                                .sum()
                                .reset_index()[response],
                                "name": f"{sensname} ({senscase})",
                                "type": "box",
                            }
                            for sensname, sensname_df in data.groupby(["SENSNAME"])
                            for senscase, senscase_df in sensname_df.groupby(
                                ["SENSCASE"]
                            )
                        ],
                        "layout": self.theme.create_themed_layout(layout),
                    },
                )
            elif plot_type == "Per sensitivity name":
                # One box per sensitivity name
                layout.update(
                    {
                        "title": "Distribution for each sensitivity name",
                        "yaxis": {"title": volume_title},
                    }
                )
                figure = wcc.Graph(
                    config={"displayModeBar": False},
                    id=self.uuid("graph"),
                    figure={
                        "data": [
                            {
                                "y": sensname_df.groupby("REAL")
                                .sum()
                                .reset_index()[response],
                                "name": f"{sensname}",
                                "type": "box",
                            }
                            for sensname, sensname_df in data.groupby(["SENSNAME"])
                        ],
                        "layout": self.theme.create_themed_layout(layout),
                    },
                )
            else:
                # Should not occur unless plot_type options are changed
                figure = html.Div("Invalid plot type")

            return figure, volume_title


def calculate_table(df: pd.DataFrame, response: str) -> Tuple[List[dict], List[dict]]:
    table: List[dict] = []
    for (sensname, senscase), dframe in df.groupby(["SENSNAME", "SENSCASE"]):
        values = dframe.groupby("REAL").sum().reset_index()[response]
        try:
            table.append(
                {
                    "Sens Name": str(sensname),
                    "Sens Case": str(senscase),
                    "Minimum": values.min(),
                    "Maximum": values.max(),
                    "Mean": values.mean(),
                    "Stddev": values.std(),
                    "P10": np.percentile(values, 90),
                    "P90": np.percentile(values, 10),
                }
            )
        except KeyError:
            pass
    columns: List[dict] = [
        {**{"name": i[0], "id": i[0]}, **i[1]}
        for i in InplaceVolumesOneByOne.TABLE_STATISTICS
    ]
    return table, columns


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_dataframe(
    dframe: pd.DataFrame,
    columns: Union[str, List[str]],
    ensemble: str,
    source: str,
    column_values: Iterable[Union[str, List[str]]],
) -> pd.DataFrame:
    df = dframe.copy()
    if not isinstance(columns, list):
        columns = [columns]
    for filt, col in zip(column_values, columns):
        if isinstance(filt, list):
            df = df.loc[df[col].isin(filt)]
        else:
            df = df.loc[df[col] == filt]
    df = df.loc[df["ENSEMBLE"] == ensemble]
    df = df.loc[df["SOURCE"] == source]
    return df


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)



import pytest

from webviz_subsurface._utils import simulation_timeseries


@pytest.mark.parametrize(
    "date,vector,interval,as_date,res",
    [
        # Testing AVG_ vectors
        ("2000-01-01", "AVG_FOPR", "monthly", False, "2000-01"),
        ("2003-05-12", "AVG_WOPR:OP_1", "monthly", False, "2003-05"),
        ("2002-05-12", "AVG_WOPR:OP_1", "yearly", False, "2002"),
        ("2002-05-12", "AVG_WOPR:OP_1", "yearly", True, "2002-01-01"),
        ("2002-05-12", "AVG_WOPR:OP_1", "daily", False, "2002-05-12"),
        ("2002-05-12", "AVG_WOPR:OP_1", "daily", True, "2002-05-12"),
        (None, "AVG_WOPR:OP_1", "daily", True, None),
        # Testing INTVL_ vectors
        ("2000-01-01", "INTVL_FOPR", "monthly", False, "2000-01"),
        ("2003-05-12", "INTVL_WGPT:OP_1", "monthly", False, "2003-05"),
        ("2002-05-12", "INTVL_WGPT:OP_1", "yearly", False, "2002"),
        ("2002-05-12", "INTVL_WGPT:OP_1", "yearly", True, "2002-01-01"),
        ("2002-05-12", "INTVL_WGPT:OP_1", "daily", False, "2002-05-12"),
        ("2002-05-12", "INTVL_WGPT:OP_1", "daily", True, "2002-05-12"),
        (None, "INTVL_WGPT:OP_1", "daily", True, None),
        # Testing vectors that are not AVG_ or INTVL_. Dates should return as is.
        ("2000-01-01", "FOPR", "monthly", False, "2000-01-01"),
        ("2003-05-12", "FGPT", "monthly", False, "2003-05-12"),
        ("2002-05-12", "WGPT:OP_1", "yearly", False, "2002-05-12"),
        ("2002-05-12", "WOPR:OP_1", "yearly", True, "2002-05-12"),
        ("2002-05-12", "TIMESTEP", "daily", False, "2002-05-12"),
        ("2002-05-12", "NSUMPROB", "daily", True, "2002-05-12"),
        (None, "WWCT:OP_2", "daily", True, None),
    ],
)
def test_date_to_interval_conversion(
    date: str, vector: str, interval: str, as_date: bool, res: str
) -> None:
    if res is None:
        assert (
            simulation_timeseries.date_to_interval_conversion(
                date=date, vector=vector, interval=interval, as_date=as_date
            )
            is None
        )
    else:
        assert (
            simulation_timeseries.date_to_interval_conversion(
                date=date, vector=vector, interval=interval, as_date=as_date
            )
            == res
        )



import io
import itertools
import json
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple

import numpy as np
import pandas as pd
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from ..._datainput.fmu_input import load_csv
from ..._datainput.well_completions import (
    get_ecl_unit_system,
    read_stratigraphy,
    read_well_attributes,
    read_well_connection_status,
    read_zone_layer_mapping,
)


class WellCompletionsDataModel:
    def __init__(
        self,
        ensemble_name: str,
        ensemble_path: str,
        compdat_file: str,
        well_connection_status_file: str,
        zone_layer_mapping_file: str,
        stratigraphy_file: str,
        well_attributes_file: str,
        kh_unit: Optional[str],
        kh_decimal_places: int,
        theme_colors: List[str],
    ) -> None:
        # pylint: disable=too-many-arguments

        self.ensemble_name = ensemble_name
        self.ensemble_path = ensemble_path
        self.compdat_file = compdat_file
        self.well_connection_status_file = well_connection_status_file
        self.zone_layer_mapping_file = zone_layer_mapping_file
        self.stratigraphy_file = stratigraphy_file
        self.well_attributes_file = well_attributes_file
        self.kh_unit = kh_unit
        self.kh_decimal_places = kh_decimal_places
        self.theme_colors = theme_colors

    def __repr__(self) -> str:
        """This is necessary for webvizstore to work"""
        return f"""
WellCompletionsDataModel {self.ensemble_name} {self.ensemble_path} {self.compdat_file}
{self.well_connection_status_file} {self.zone_layer_mapping_file} {self.stratigraphy_file}
{self.well_attributes_file}
        """

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    @webvizstore
    def create_ensemble_dataset(self) -> io.BytesIO:
        """Creates the well completion data set for the WellCompletions component

        Returns a dictionary on a given format specified here:
        https://github.com/equinor/webviz-subsurface-components/blob/master/react/src/lib/inputSchema/wellCompletions.json
        """
        df = load_csv(
            ensemble_paths={self.ensemble_name: self.ensemble_path},
            csv_file=self.compdat_file,
        )
        df = df[["REAL", "DATE", "WELL", "I", "J", "K1", "OP/SH", "KH"]]
        df.DATE = pd.to_datetime(df.DATE).dt.date

        df_connstatus = read_well_connection_status(
            ensemble_path=self.ensemble_path,
            well_connection_status_file=self.well_connection_status_file,
        )
        df_zone_layer = read_zone_layer_mapping(
            ensemble_path=self.ensemble_path,
            zone_layer_mapping_file=self.zone_layer_mapping_file,
        )
        stratigraphy = read_stratigraphy(
            ensemble_path=self.ensemble_path, stratigraphy_file=self.stratigraphy_file
        )
        well_attributes = read_well_attributes(
            ensemble_path=self.ensemble_path,
            well_attributes_file=self.well_attributes_file,
        )
        if self.kh_unit is None:
            self.kh_unit, self.kh_decimal_places = get_kh_unit(
                ensemble_path=self.ensemble_path
            )

        if df_connstatus is not None:
            df = merge_compdat_and_connstatus(df, df_connstatus)

        time_steps = sorted(df.DATE.unique())
        realizations = list(sorted(df.REAL.unique()))

        if df_zone_layer.empty:
            if stratigraphy is None:
                df["ZONE"] = df.agg(lambda x: f"Layer {x['K1']}", axis=1)
                df["COLOR"] = np.nan
            else:
                raise ValueError(
                    "It is not permitted to define the stratigraphy, but not the "
                    "zone ➔ layer mapping. If neither input is provided then layers "
                    "will be used as zones (NB! this can be slow with many wells and realizations)"
                )
        else:
            reals_without_mapping = set(df["REAL"].unique()) - set(
                df_zone_layer["REAL"].unique()
            )
            if reals_without_mapping:
                raise ValueError(
                    "The zone ➔ layer mapping seems to be missing for the "
                    f"following realizations: {reals_without_mapping}"
                )

            # We know that all realizations has a zone layer mapping, but we do
            # not require that all layers are defined in the zone layer mapping.
            # Layers missing from the zone layer mapping will be filtered out here.
            df = df.merge(df_zone_layer, on=["K1", "REAL"], how="inner")

        # Get the zone names in the correct order
        zone_names = (
            df.drop_duplicates(subset=["K1", "ZONE"])
            .sort_values(by="K1")["ZONE"]
            .unique()
        )

        zone_color_mapping = {
            item["ZONE"]: item["COLOR"]
            for item in df[["ZONE", "COLOR"]]
            .dropna()
            .drop_duplicates(keep="first")
            .to_dict("records")
        }

        result = {
            "version": "1.1.0",
            "units": {
                "kh": {"unit": self.kh_unit, "decimalPlaces": self.kh_decimal_places}
            },
            "stratigraphy": extract_stratigraphy(
                zone_names, stratigraphy, zone_color_mapping, self.theme_colors
            ),
            "timeSteps": [str(dte) for dte in time_steps],
            "wells": extract_wells(
                df, zone_names, time_steps, realizations, well_attributes
            ),
        }
        return io.BytesIO(json.dumps(result).encode())

    @property
    def webviz_store(self) -> Tuple[Callable, List[Dict]]:
        return (
            self.create_ensemble_dataset,
            [
                {
                    "self": self,
                }
            ],
        )


def merge_compdat_and_connstatus(
    df_compdat: pd.DataFrame, df_connstatus: pd.DataFrame
) -> pd.DataFrame:
    """This function merges the compdat data (exported with ecl2df) with the well connection
    status data (extracted from the CPI summary data). The connection status data will
    be used for wells where it exists. The KH will be merged from the compdat. For wells
    that are not in the connection status data, the compdat data will be used as it is.

    This approach is fast, but a couple of things should be noted:
    * in the connection status data, a connection is not set to SHUT before it has been OPEN. \
    In the compdat data, some times all connections are first defined and the opened later.
    * any connections that are in compdat but not in connections status will be ignored \
    (e.g. connections that are always shut)
    * there is no logic to handle KH changing with time for the same connection (this \
    can easily be added using apply in pandas, but it is very rare and slows down the function
    significantly)
    * if connection status is missing for a realization, but compdat exists, compdat will also \
    be ignored.
    """
    match_on = ["REAL", "WELL", "I", "J", "K1"]
    df = pd.merge(df_connstatus, df_compdat[match_on + ["KH"]], on=match_on, how="left")

    # There will often be several rows (with different OP/SH) matching in compdat.
    # Only the first is kept
    df.drop_duplicates(subset=["DATE"] + match_on, keep="first", inplace=True)

    # Concat from compdat the wells that are not in well connection status
    df = pd.concat([df, df_compdat[~df_compdat.WELL.isin(df.WELL.unique())]])

    df = df.reset_index(drop=True)
    df.KH = df.KH.fillna(0)
    return df


def get_kh_unit(ensemble_path: str) -> Tuple[str, int]:
    """Returns kh unit and decimal places based on the unit system of the eclipse deck"""
    units = {
        "METRIC": ("mD·m", 2),
        "FIELD": ("mD·ft", 2),
        "LAB": ("mD·cm", 2),
        "PVT-M": ("mD·m", 2),
    }
    unit_system = get_ecl_unit_system(ensemble_path=ensemble_path)
    if unit_system is not None:
        return units[unit_system]
    return ("", 2)


def get_time_series(df: pd.DataFrame, time_steps: list) -> tuple:
    """Create two lists with values for each time step
    * the first one is on the form [0,0,0,1,1,1,1,-1,-1,-1] where '0' means no event,\
    '1' is open, '-1' is shut.
    * the second is with sum of kh values for the open compdats in each zone

    The input data frame is assumed to contain data for single well,
    single zone and single realization.
    """
    if df.empty:
        return [0] * len(time_steps), [0] * len(time_steps)

    events, kh_values = [], []
    event_value, kh_value = 0, 0
    unique_dates = df.DATE.unique()

    for timestep in time_steps:
        if timestep in unique_dates:
            df_timestep = df[df.DATE == timestep]
            df_timestep_open = df_timestep[df_timestep["OP/SH"] == "OPEN"]

            # if minimum one of the compdats for the zone is OPEN then the zone is considered open
            event_value = 1 if not df_timestep_open.empty else -1
            kh_value = df_timestep_open.KH.sum()

        events.append(event_value)
        kh_values.append(kh_value)
    return events, kh_values


def get_completion_events_and_kh(
    df: pd.DataFrame, zone_names: list, time_steps: list
) -> tuple:
    """Extracts completion events ad kh values into two lists of lists of lists,
    one with completions events and one with kh values
    * Axis 0 is realization
    * Axis 1 is zone
    * Axis 2 is time step
    """
    compl_events, kh_values = [], []
    for _, realdata in df.groupby("REAL"):
        compl_events_real, kh_real = [], []
        for zone_name in zone_names:
            zone_data = realdata[realdata.ZONE == zone_name]
            compl_events_real_zone, kh_real_zone = get_time_series(
                zone_data, time_steps
            )
            compl_events_real.append(compl_events_real_zone)
            kh_real.append(kh_real_zone)

        compl_events.append(compl_events_real)
        kh_values.append(kh_real)
    return compl_events, kh_values


def format_time_series(
    open_frac: List[float],
    shut_frac: List[float],
    kh_mean: List[float],
    kh_min: List[float],
    kh_max: List[float],
) -> Optional[Dict]:
    """The functions takes in five lists with values per timestep
    * fractions of realizations open in this zone
    * fractions of realizations shut in this zone
    * kh mean over open realizations
    * kh min over open realizations
    * kh max over open realizations

    Returns the data in compact form:
    {
        t: [3, 5],
        open: [0.25, 1.0],
        shut: [0.75, 0.0],
        khMean: [500, 1000],
        khMin: [200, 800],
        khMean: [600, 1500]
    }
    """
    output: Dict[str, list] = {
        "t": [],
        "open": [],
        "shut": [],
        "khMean": [],
        "khMin": [],
        "khMax": [],
    }
    prev_open_val = prev_shut_val = 0.0

    for i, _ in enumerate(open_frac):
        open_val = open_frac[i]
        shut_val = shut_frac[i]
        if open_val != prev_open_val or shut_val != prev_shut_val:
            output["t"].append(i)
            output["open"].append(open_val)
            output["shut"].append(shut_val)
            output["khMean"].append(kh_mean[i])
            output["khMin"].append(kh_min[i])
            output["khMax"].append(kh_max[i])
        prev_open_val = open_val
        prev_shut_val = shut_val
    return output


def calc_over_realizations(
    compl_events: list, kh_values: list, realizations: list
) -> tuple:
    """Takes in two three dimensional lists where the levels are: 1. realization \
    2. zones and 3. timesteps

    Returns two dimensional lists where calculations have been done over the \
    realization level.
    """
    # calculate fraction of open realizations
    open_count = np.maximum(np.asarray(compl_events), 0)  # remove -1
    open_count_reduced = open_count.sum(axis=0)  # sum over realizations
    open_frac = (
        np.asarray(open_count_reduced, dtype=np.float64) / float(len(realizations))
    ).round(decimals=3)

    # calculate fraction of shut realizations
    shut_count = (
        np.minimum(np.asarray(compl_events), 0) * -1
    )  # remove +1 and convert -1 to 1
    shut_count_reduced = shut_count.sum(axis=0)  # sum over realizations
    shut_frac = (
        np.asarray(shut_count_reduced, dtype=np.float64) / float(len(realizations))
    ).round(decimals=3)

    # calculate khMean, khMin and khMax
    np_kh_values = np.asarray(kh_values)
    kh_mean = (np_kh_values.sum(axis=0) / float(len(realizations))).round(decimals=2)
    kh_min = np_kh_values.min(axis=0).round(decimals=2)
    kh_max = np_kh_values.max(axis=0).round(decimals=2)

    return open_frac, shut_frac, kh_mean, kh_min, kh_max


def extract_well(
    df: pd.DataFrame, well: str, zone_names: list, time_steps: list, realizations: list
) -> Dict[str, Any]:
    # pylint: disable=too-many-locals
    """Extract completion events and kh values for a single well"""
    well_dict: Dict[str, Any] = {}
    well_dict["name"] = well

    compl_events, kh_values = get_completion_events_and_kh(df, zone_names, time_steps)
    open_frac, shut_frac, kh_mean, kh_min, kh_max = calc_over_realizations(
        compl_events, kh_values, realizations
    )
    result = {}
    for (
        zone_name,
        open_frac_zone,
        shut_frac_zone,
        kh_mean_zone,
        kh_min_zone,
        kh_max_zone,
    ) in zip(zone_names, open_frac, shut_frac, kh_mean, kh_min, kh_max):
        if sum(open_frac_zone) != 0.0 or sum(shut_frac_zone) != 0.0:
            result[zone_name] = format_time_series(
                open_frac_zone, shut_frac_zone, kh_mean_zone, kh_min_zone, kh_max_zone
            )
    well_dict["completions"] = result
    return well_dict


def extract_wells(
    df: pd.DataFrame,
    zone_names: list,
    time_steps: list,
    realizations: list,
    well_attributes: Optional[dict],
) -> List[Dict]:
    """Generates the wells part of the input dictionary to the WellCompletions component"""
    well_list = []
    for well_name, well_group in df.groupby("WELL"):
        well_data = extract_well(
            well_group, well_name, zone_names, time_steps, realizations
        )
        well_data["attributes"] = (
            well_attributes[well_name]
            if (well_attributes is not None and well_name in well_attributes)
            else {}
        )
        well_list.append(well_data)
    return well_list


def add_colors_to_stratigraphy(
    stratigraphy: List[Dict[str, Any]],
    zone_color_mapping: Optional[Dict[str, str]],
    color_iterator: Iterator,
) -> List[Dict[str, Any]]:
    """Add colors to the stratigraphy tree. The function will recursively parse the tree.

    There are tree sources of color:
    1. The color is given in the stratigraphy list, in which case nothing is done to the node
    2. The color is given in the lyr file, and passed to this function in the zone->color map
    3. If none of the above applies, the color will be taken from the theme color iterable for \
    the leaves. For other levels, a dummy color grey is used
    """
    for zonedict in stratigraphy:
        if "color" not in zonedict:
            if (
                zone_color_mapping is not None
                and zonedict["name"] in zone_color_mapping
            ):
                zonedict["color"] = zone_color_mapping[zonedict["name"]]
            elif "subzones" not in zonedict:
                zonedict["color"] = next(
                    color_iterator
                )  # theme colors only applied on leaves
            else:
                zonedict["color"] = "#808080"  # grey
        if "subzones" in zonedict:
            zonedict["subzones"] = add_colors_to_stratigraphy(
                zonedict["subzones"], zone_color_mapping, color_iterator
            )
    return stratigraphy


def filter_valid_nodes(
    stratigraphy: List[Dict[str, Any]], valid_zone_names: list
) -> Tuple[List, List]:
    """Returns the stratigraphy tree with only valid nodes.
    A node is considered valid if it self or one of it's subzones are in the
    valid zone names list (passed from the lyr file)

    The function recursively parses the tree to add valid nodes.
    """
    output = []
    remaining_valid_zones = valid_zone_names
    for zonedict in stratigraphy:
        if "subzones" in zonedict:
            zonedict["subzones"], remaining_valid_zones = filter_valid_nodes(
                zonedict["subzones"], remaining_valid_zones
            )
        if zonedict["name"] in remaining_valid_zones:
            if "subzones" in zonedict and not zonedict["subzones"]:
                zonedict.pop("subzones")
            output.append(zonedict)
            remaining_valid_zones = [
                zone for zone in remaining_valid_zones if zone != zonedict["name"]
            ]  # remove zone name from valid zones if it is found in the stratigraphy
        elif "subzones" in zonedict and zonedict["subzones"]:
            output.append(zonedict)

    return output, remaining_valid_zones


def extract_stratigraphy(
    zone_names: List[str],
    stratigraphy: Optional[List[Dict[str, Any]]],
    zone_color_mapping: Optional[Dict[str, str]],
    theme_colors: list,
) -> List[Dict[str, Any]]:
    """Returns the stratigraphy part of the data set"""
    color_iterator = itertools.cycle(theme_colors)

    if stratigraphy is None:
        return [
            {
                "name": zone,
                "color": zone_color_mapping[zone]
                if zone_color_mapping is not None and zone in zone_color_mapping
                else next(color_iterator),
            }
            for zone in zone_names
        ]

    # If stratigraphy is not None the following is done:
    stratigraphy, remaining_valid_zones = filter_valid_nodes(stratigraphy, zone_names)

    if remaining_valid_zones:
        raise ValueError(
            "The following zones are defined in the zone ➔ layer mapping, "
            f"but not in the stratigraphy: {remaining_valid_zones}"
        )

    # Zones not found in the stratigraphy is added to the end.
    for zone_name in remaining_valid_zones:
        stratigraphy.append({"name": zone_name})

    return add_colors_to_stratigraphy(stratigraphy, zone_color_mapping, color_iterator)



import numpy as np
import xtgeo
from webviz_config.common_cache import CACHE


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def load_surface(surface_path: str) -> xtgeo.RegularSurface:
    return xtgeo.surface_from_file(surface_path)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_surface_fence(fence: np.ndarray, surface: xtgeo.RegularSurface) -> np.ndarray:
    return surface.get_fence(fence)



from typing import Any, Dict, List, Tuple

from dash import Input, Output, State, callback, dcc
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import ViewABC

from ..._types import PressureType, VfpParam
from ..._utils import VfpDataModel, VfpTable
from ._settings import Filters, PressureOption, Selections, Vizualisation
from ._utils import VfpFigureBuilder
from ._view_elements import VfpGraph


class VfpView(ViewABC):
    class Ids(StrEnum):
        VFP_GRAPH = "vfp-graph"
        VFP_METADATA = "vfp-metadata"
        SELECTIONS = "selections"
        PRESSURE_OPTION = "pressure-option"
        VIZUALISATION = "vizualisation"
        FILTERS = "filters"

    def __init__(self, data_model: VfpDataModel) -> None:
        super().__init__("VFP Analysis")

        self._data_model = data_model

        self.add_settings_group(
            Selections(self._data_model.vfp_names), VfpView.Ids.SELECTIONS
        )
        self.add_settings_group(PressureOption(), VfpView.Ids.PRESSURE_OPTION)
        self.add_settings_group(
            Vizualisation(self._data_model.vfp_names), VfpView.Ids.VIZUALISATION
        )
        self.add_settings_group(Filters(), VfpView.Ids.FILTERS)
        column = self.add_column()
        column.add_view_element(VfpGraph(), VfpView.Ids.VFP_GRAPH)
        # column.add_view_element(VfpMetadata(), VfpView.Ids.VFP_METADATA)

    def set_callbacks(self) -> None:
        @callback(
            # Options
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.THP),
                "options",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.WFR),
                "options",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.GFR),
                "options",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.ALQ),
                "options",
            ),
            # Values
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.THP),
                "value",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.WFR),
                "value",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.GFR),
                "value",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.ALQ),
                "value",
            ),
            # Labels
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.THP_LABEL),
                "children",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.WFR_LABEL),
                "children",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.GFR_LABEL),
                "children",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.ALQ_LABEL),
                "children",
            ),
            # Size of filter boxes
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.THP),
                "size",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.WFR),
                "size",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.GFR),
                "size",
            ),
            Output(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.ALQ),
                "size",
            ),
            # Table metadata
            Output(
                self.settings_group_unique_id(
                    self.Ids.SELECTIONS, Selections.Ids.METADATA_DIALOG
                ),
                "children",
            ),
            # Input
            Input(
                self.settings_group_unique_id(
                    self.Ids.SELECTIONS, Selections.Ids.VFP_NAME
                ),
                "value",
            ),
        )
        @callback_typecheck
        def _update_filters(
            vfp_name: str,
        ) -> Tuple[
            List[Dict[str, float]],
            List[Dict[str, float]],
            List[Dict[str, float]],
            List[Dict[str, float]],
            List[float],
            List[float],
            List[float],
            List[float],
            str,
            str,
            str,
            str,
            int,
            int,
            int,
            int,
            dcc.Markdown,
        ]:
            """Updates the filter values, sets the initial selection, and
            updates the filter label.
            """
            vfp_table: VfpTable = self._data_model.get_vfp_table(vfp_name)

            thp_dict = vfp_table.params[VfpParam.THP]
            wfr_dict = vfp_table.params[VfpParam.WFR]
            gfr_dict = vfp_table.params[VfpParam.GFR]
            alq_dict = vfp_table.params[VfpParam.ALQ]

            return (
                [{"label": value, "value": idx} for idx, value in thp_dict.items()],
                [{"label": value, "value": idx} for idx, value in wfr_dict.items()],
                [{"label": value, "value": idx} for idx, value in gfr_dict.items()],
                [{"label": value, "value": idx} for idx, value in alq_dict.items()],
                list(thp_dict.keys()),
                [list(wfr_dict.keys())[0]],
                [list(gfr_dict.keys())[0]],
                [list(alq_dict.keys())[0]],
                f"THP: {vfp_table.param_types[VfpParam.THP].name}",
                f"WFR: {vfp_table.param_types[VfpParam.WFR].name}",
                f"GFR: {vfp_table.param_types[VfpParam.GFR].name}",
                f"ALQ: {vfp_table.param_types[VfpParam.ALQ].name}",
                min(6, len(thp_dict)),
                min(6, len(wfr_dict)),
                min(6, len(gfr_dict)),
                min(6, len(alq_dict)),
                dcc.Markdown(vfp_table.get_metadata_markdown()),
            )

        @callback(
            # Options
            Output(
                self.view_element_unique_id(self.Ids.VFP_GRAPH, VfpGraph.Ids.GRAPH),
                "figure",
            ),
            Input(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.THP),
                "value",
            ),
            Input(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.WFR),
                "value",
            ),
            Input(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.GFR),
                "value",
            ),
            Input(
                self.settings_group_unique_id(self.Ids.FILTERS, Filters.Ids.ALQ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.VIZUALISATION, Vizualisation.Ids.COLOR_BY
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.PRESSURE_OPTION, PressureOption.Ids.PRESSURE_OPTION
                ),
                "value",
            ),
            State(
                self.settings_group_unique_id(
                    self.Ids.SELECTIONS, Selections.Ids.VFP_NAME
                ),
                "value",
            ),
        )
        @callback_typecheck
        def _update_vfp_graph(
            thps: List[int],
            wfrs: List[int],
            gfrs: List[int],
            alqs: List[int],
            color_by: VfpParam,
            pressure_type: PressureType,
            vfp_name: str,
        ) -> Dict[str, Any]:
            # pylint: disable=too-many-locals
            vfp_table = self._data_model.get_vfp_table(vfp_name)

            figure_builder = VfpFigureBuilder(vfp_name=vfp_name)

            selected_indices = {
                VfpParam.THP: thps,
                VfpParam.WFR: wfrs,
                VfpParam.GFR: gfrs,
                VfpParam.ALQ: alqs,
            }
            selected_color_by_values = vfp_table.get_values(
                color_by, selected_indices[color_by]
            )
            cmax = max(selected_color_by_values)
            cmin = min(selected_color_by_values)
            showscale = True
            for thp_idx in thps:
                for wfr_idx in wfrs:
                    for gfr_idx in gfrs:
                        for alq_idx in alqs:
                            figure_builder.add_vfp_curve(
                                rates=vfp_table.get_values(vfp_param=VfpParam.RATE),
                                bhp_values=vfp_table.get_bhp_series(
                                    pressure_type, thp_idx, wfr_idx, gfr_idx, alq_idx
                                ),
                                cmax=cmax,
                                cmin=cmin,
                                vfp_table=vfp_table,
                                indices={
                                    VfpParam.THP: thp_idx,
                                    VfpParam.WFR: wfr_idx,
                                    VfpParam.GFR: gfr_idx,
                                    VfpParam.ALQ: alq_idx,
                                },
                                color_by=color_by,
                                showscale=showscale,
                            )
                            if showscale:
                                showscale = False

            figure_builder.set_xaxis_settings(title=vfp_table.get_rate_label())
            figure_builder.set_yaxis_settings(
                title=vfp_table.get_bhp_label(pressure_type=pressure_type)
            )

            return figure_builder.get_figure()



import logging
import shutil
from enum import Enum
from pathlib import Path
from typing import List, Optional

import numpy as np
import pandas as pd
import xtgeo

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._roff_file_discovery import GridFileInfo, GridParameterFileInfo
from .ensemble_grid_provider import EnsembleGridProvider

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class Col:
    TYPE = "type"
    REAL = "real"
    ATTRIBUTE = "attribute"
    NAME = "name"
    DATESTR = "datestr"
    ORIGINAL_PATH = "original_path"
    REL_PATH = "rel_path"


class GridType(str, Enum):
    GEOMETRY = "geometry"
    STATIC_PROPERTY = "static_property"
    DYNAMIC_PROPERTY = "dynamic_property"


class ProviderImplRoff(EnsembleGridProvider):
    def __init__(
        self, provider_id: str, provider_dir: Path, grid_inventory_df: pd.DataFrame
    ) -> None:
        self._provider_id = provider_id
        self._provider_dir = provider_dir
        self._inventory_df = grid_inventory_df

    @staticmethod
    # pylint: disable=too-many-locals
    def write_backing_store(
        storage_dir: Path,
        storage_key: str,
        grid_geometries_info: List[GridFileInfo],
        grid_parameters_info: List[GridParameterFileInfo],
        avoid_copying_grid_data: bool,
    ) -> None:
        """If avoid_copying_grid_data if True, the specified grid data will NOT be copied
        into the backing store, but will be referenced from their source locations.
        Note that this is only useful when running in non-portable mode and will fail
        in portable mode.
        """

        timer = PerfTimer()

        do_copy_grid_data_into_store = not avoid_copying_grid_data

        # All data for this provider will be stored inside a sub-directory
        # given by the storage key
        provider_dir = storage_dir / storage_key
        LOGGER.info(f"Writing grid data backing store to: {provider_dir}")
        provider_dir.mkdir(parents=True, exist_ok=True)

        type_arr: List[GridType] = []
        real_arr: List[int] = []
        attribute_arr: List[str] = []
        name_arr: List[str] = []
        datestr_arr: List[str] = []
        rel_path_arr: List[str] = []
        original_path_arr: List[str] = []
        for grid_info in grid_geometries_info:
            type_arr.append(GridType.GEOMETRY)
            name_arr.append(grid_info.name)
            real_arr.append(grid_info.real)
            attribute_arr.append("")
            datestr_arr.append("")
            original_path_arr.append(grid_info.path)
            rel_path_in_store = ""

            if do_copy_grid_data_into_store:
                rel_path_in_store = _compose_rel_grid_pathstr(
                    real=grid_info.real,
                    attribute=None,
                    name=grid_info.name,
                    datestr=None,
                    extension=Path(grid_info.path).suffix,
                )

            rel_path_arr.append(rel_path_in_store)

        for grid_parameter_info in grid_parameters_info:
            name_arr.append(grid_parameter_info.name)
            real_arr.append(grid_parameter_info.real)
            attribute_arr.append(grid_parameter_info.attribute)
            if grid_parameter_info.datestr:
                datestr_arr.append(grid_parameter_info.datestr)
                type_arr.append(GridType.DYNAMIC_PROPERTY)
            else:
                datestr_arr.append("")
                type_arr.append(GridType.STATIC_PROPERTY)

            original_path_arr.append(grid_parameter_info.path)

            rel_path_in_store = ""
            if do_copy_grid_data_into_store:
                rel_path_in_store = _compose_rel_grid_pathstr(
                    real=grid_parameter_info.real,
                    attribute=grid_parameter_info.attribute,
                    name=grid_parameter_info.name,
                    datestr=grid_parameter_info.datestr,
                    extension=Path(grid_parameter_info.path).suffix,
                )

            rel_path_arr.append(rel_path_in_store)

        timer.lap_s()
        if do_copy_grid_data_into_store:
            LOGGER.debug(
                f"Copying {len(original_path_arr)} grid data into backing store..."
            )
            _copy_grid_parameters_into_provider_dir(
                original_path_arr, rel_path_arr, provider_dir
            )
        et_copy_s = timer.lap_s()

        grid_inventory_df = pd.DataFrame(
            {
                Col.TYPE: type_arr,
                Col.REAL: real_arr,
                Col.ATTRIBUTE: attribute_arr,
                Col.NAME: name_arr,
                Col.DATESTR: datestr_arr,
                Col.REL_PATH: rel_path_arr,
                Col.ORIGINAL_PATH: original_path_arr,
            }
        )

        parquet_file_name = provider_dir / "grid_inventory.parquet"
        grid_inventory_df.to_parquet(path=parquet_file_name)

        if do_copy_grid_data_into_store:
            LOGGER.debug(
                f"Wrote grid backing store in: {timer.elapsed_s():.2f}s ("
                f"copy={et_copy_s:.2f}s)"
            )
        else:
            LOGGER.debug(
                f"Wrote grid backing store without copying grid data in: "
                f"{timer.elapsed_s():.2f}s"
            )

    @staticmethod
    def from_backing_store(
        storage_dir: Path,
        storage_key: str,
    ) -> Optional["ProviderImplRoff"]:

        provider_dir = storage_dir / storage_key
        parquet_file_name = provider_dir / "grid_inventory.parquet"

        try:
            grid_inventory_df = pd.read_parquet(path=parquet_file_name)
            return ProviderImplRoff(storage_key, provider_dir, grid_inventory_df)
        except FileNotFoundError:
            return None

    def provider_id(self) -> str:
        return self._provider_id

    def static_property_names(self) -> List[str]:
        return sorted(
            list(
                self._inventory_df.loc[
                    self._inventory_df[Col.TYPE] == GridType.STATIC_PROPERTY
                ][Col.ATTRIBUTE].unique()
            )
        )

    def dynamic_property_names(self) -> List[str]:
        return sorted(
            list(
                self._inventory_df.loc[
                    self._inventory_df[Col.TYPE] == GridType.DYNAMIC_PROPERTY
                ][Col.ATTRIBUTE].unique()
            )
        )

    def dates_for_dynamic_property(self, property_name: str) -> Optional[List[str]]:
        dates = sorted(
            list(
                self._inventory_df.loc[
                    (self._inventory_df[Col.TYPE] == GridType.DYNAMIC_PROPERTY)
                    & (self._inventory_df[Col.ATTRIBUTE] == property_name)
                ][Col.DATESTR].unique()
            )
        )
        if len(dates) == 1 and not bool(dates[0]):
            return None

        return dates

    def realizations(self) -> List[int]:
        unique_reals = self._inventory_df[Col.REAL].unique()

        # Sort and strip out any entries with real == -1
        return sorted([r for r in unique_reals if r >= 0])

    def get_3dgrid(self, realization: int) -> xtgeo.Grid:
        df = self._inventory_df.loc[self._inventory_df[Col.TYPE] == GridType.GEOMETRY]
        df = df.loc[df[Col.REAL] == realization]

        df = df[[Col.REL_PATH, Col.ORIGINAL_PATH]]
        fn_list: List[str] = []
        for _index, row in df.iterrows():
            if row[Col.REL_PATH]:
                fn_list.append(self._provider_dir / row[Col.REL_PATH])
            else:
                fn_list.append(row[Col.ORIGINAL_PATH])
        if len(fn_list) == 0:
            LOGGER.warning(f"No grid geometry found for realization {realization}")
            return None
        if len(fn_list) > 1:
            raise ValueError(
                f"Multiple grid geometries found for: {realization}"
                "Something has gone terribly wrong."
            )

        grid = xtgeo.grid_from_file(fn_list[0])
        return grid

    def get_static_property_values(
        self, property_name: str, realization: int
    ) -> Optional[np.ndarray]:
        fn_list: List[str] = self._locate_static_property(
            property_name=property_name, realizations=[realization]
        )
        if len(fn_list) == 0:
            LOGGER.warning(f"No grid parameter found for realization {realization}")
            return None
        if len(fn_list) > 1:
            raise ValueError(
                f"Multiple grid parameters found for: {realization}"
                "Something has gone terribly wrong."
            )
        grid_property = xtgeo.gridproperty_from_file(fn_list[0])
        fill_value = np.nan if not grid_property.isdiscrete else -1
        return grid_property.get_npvalues1d(order="F", fill_value=fill_value).ravel()

    def get_dynamic_property_values(
        self, property_name: str, property_date: str, realization: int
    ) -> Optional[np.ndarray]:
        fn_list: List[str] = self._locate_dynamic_property(
            property_name=property_name,
            property_datestr=property_date,
            realizations=[realization],
        )
        if len(fn_list) == 0:
            LOGGER.warning(f"No grid parameter found for realization {realization}")
            return None
        if len(fn_list) > 1:
            raise ValueError(
                f"Multiple grid parameters found for: {realization}"
                "Something has gone terribly wrong."
            )
        grid_property = xtgeo.gridproperty_from_file(fn_list[0])
        return grid_property.get_npvalues1d(order="F").ravel()

    def _locate_static_property(
        self, property_name: str, realizations: List[int]
    ) -> List[str]:
        """Returns list of file names matching the specified filter criteria"""
        df = self._inventory_df.loc[
            self._inventory_df[Col.TYPE] == GridType.STATIC_PROPERTY
        ]

        df = df.loc[
            (df[Col.ATTRIBUTE] == property_name) & (df[Col.REAL].isin(realizations))
        ]

        df = df[[Col.REL_PATH, Col.ORIGINAL_PATH]]

        # Return file name within backing store if the data was copied there,
        # otherwise return the original source file name
        fn_list: List[str] = []
        for _index, row in df.iterrows():
            if row[Col.REL_PATH]:
                fn_list.append(self._provider_dir / row[Col.REL_PATH])
            else:
                fn_list.append(row[Col.ORIGINAL_PATH])

        return fn_list

    def _locate_dynamic_property(
        self, property_name: str, property_datestr: str, realizations: List[int]
    ) -> List[str]:
        """Returns list of file names matching the specified filter criteria"""
        df = self._inventory_df.loc[
            self._inventory_df[Col.TYPE] == GridType.DYNAMIC_PROPERTY
        ]

        df = df.loc[
            (df[Col.ATTRIBUTE] == property_name)
            & (df[Col.DATESTR] == property_datestr)
            & (df[Col.REAL].isin(realizations))
        ]

        df = df[[Col.REL_PATH, Col.ORIGINAL_PATH]]

        # Return file name within backing store if the data was copied there,
        # otherwise return the original source file name
        fn_list: List[str] = []
        for _index, row in df.iterrows():
            if row[Col.REL_PATH]:
                fn_list.append(self._provider_dir / row[Col.REL_PATH])
            else:
                fn_list.append(row[Col.ORIGINAL_PATH])

        return fn_list


def _copy_grid_parameters_into_provider_dir(
    original_path_arr: List[str],
    rel_path_arr: List[str],
    provider_dir: Path,
) -> None:
    for src_path, dst_rel_path in zip(original_path_arr, rel_path_arr):
        shutil.copyfile(src_path, provider_dir / dst_rel_path)

    # full_dst_path_arr = [storage_dir / dst_rel_path for dst_rel_path in store_path_arr]
    # with ProcessPoolExecutor() as executor:
    #     executor.map(shutil.copyfile, original_path_arr, full_dst_path_arr)


def _compose_rel_grid_pathstr(
    real: int,
    name: str,
    attribute: Optional[str],
    datestr: Optional[str],
    extension: str,
) -> str:
    """Compose path to grid file, relative to provider's directory"""
    if not attribute and not datestr:
        return str(Path(f"{real}--{name}{extension}"))
    if not datestr:
        return str(Path(f"{real}--{name}--{attribute}{extension}"))
    return str(Path(f"{real}--{name}--{attribute}--{datestr}{extension}"))



from typing import List

import webviz_core_components as wcc
from dash import Input, Output, callback, html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from .._types import (
    FanchartOptions,
    StatisticsOptions,
    TraceOptions,
    VisualizationOptions,
)


class VisualizationSettings(SettingsGroupABC):
    class Ids(StrEnum):
        VISUALIZATION_RADIO_ITEMS = "visualization-radio-items"
        PLOT_TRACE_OPTIONS_CHECKLIST = "plot-trace-options-checklist"
        PLOT_STATISTICS_OPTIONS_CHECKLIST = "plot-statistics-options-checklist"
        PLOT_FANCHART_OPTIONS_CHECKLIST = "plot-fanchart-options-checklist"
        PLOT_OPTIONS = "plot-options"

    def __init__(self, selected_visualization: VisualizationOptions) -> None:
        super().__init__("Visualization")
        self._selected_visualization = selected_visualization

    def layout(self) -> List[Component]:
        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(
                    VisualizationSettings.Ids.VISUALIZATION_RADIO_ITEMS
                ),
                options=[
                    {
                        "label": "Individual realizations",
                        "value": VisualizationOptions.REALIZATIONS,
                    },
                    {
                        "label": "Statistical lines",
                        "value": VisualizationOptions.STATISTICS,
                    },
                    {
                        "label": "Statistical fanchart",
                        "value": VisualizationOptions.FANCHART,
                    },
                    {
                        "label": "Statistics + Realizations",
                        "value": VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                    },
                ],
                value=self._selected_visualization,
            ),
            wcc.LabeledContainer(
                label="Options",
                id=self.register_component_unique_id(
                    VisualizationSettings.Ids.PLOT_OPTIONS
                ),
                children=self._plot_options_layout(
                    selected_visualization=self._selected_visualization,
                ),
            ),
        ]

    def _plot_options_layout(
        self,
        selected_visualization: VisualizationOptions,
    ) -> html.Div:
        return html.Div(
            children=[
                wcc.Checklist(
                    id=self.register_component_unique_id(
                        VisualizationSettings.Ids.PLOT_TRACE_OPTIONS_CHECKLIST
                    ),
                    style={"display": "block"},
                    options=[
                        {"label": "History", "value": TraceOptions.HISTORY},
                        {
                            "label": "Observation",
                            "value": TraceOptions.OBSERVATIONS,
                        },
                    ],
                    value=[TraceOptions.HISTORY, TraceOptions.OBSERVATIONS],
                ),
                wcc.Checklist(
                    id=self.register_component_unique_id(
                        VisualizationSettings.Ids.PLOT_STATISTICS_OPTIONS_CHECKLIST
                    ),
                    style={"display": "block"}
                    if selected_visualization
                    in [
                        VisualizationOptions.STATISTICS,
                        VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                    ]
                    else {"display": "none"},
                    options=[
                        {"label": "Mean", "value": StatisticsOptions.MEAN},
                        {
                            "label": "P10 (high)",
                            "value": StatisticsOptions.P10,
                        },
                        {
                            "label": "P50 (median)",
                            "value": StatisticsOptions.P50,
                        },
                        {
                            "label": "P90 (low)",
                            "value": StatisticsOptions.P90,
                        },
                        {"label": "Maximum", "value": StatisticsOptions.MAX},
                        {"label": "Minimum", "value": StatisticsOptions.MIN},
                    ],
                    value=[
                        StatisticsOptions.MEAN,
                        StatisticsOptions.P10,
                        StatisticsOptions.P90,
                    ],
                ),
                wcc.Checklist(
                    id=self.register_component_unique_id(
                        VisualizationSettings.Ids.PLOT_FANCHART_OPTIONS_CHECKLIST
                    ),
                    style={"display": "block"}
                    if VisualizationOptions.FANCHART == selected_visualization
                    else {"display": "none"},
                    options=[
                        {
                            "label": FanchartOptions.MEAN,
                            "value": FanchartOptions.MEAN,
                        },
                        {
                            "label": FanchartOptions.P10_P90,
                            "value": FanchartOptions.P10_P90,
                        },
                        {
                            "label": FanchartOptions.MIN_MAX,
                            "value": FanchartOptions.MIN_MAX,
                        },
                    ],
                    value=[
                        FanchartOptions.MEAN,
                        FanchartOptions.P10_P90,
                        FanchartOptions.MIN_MAX,
                    ],
                ),
            ],
        )

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(
                    VisualizationSettings.Ids.PLOT_STATISTICS_OPTIONS_CHECKLIST
                ).to_string(),
                "style",
            ),
            Output(
                self.component_unique_id(
                    VisualizationSettings.Ids.PLOT_FANCHART_OPTIONS_CHECKLIST
                ).to_string(),
                "style",
            ),
            Input(
                self.component_unique_id(
                    VisualizationSettings.Ids.VISUALIZATION_RADIO_ITEMS
                ).to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _update_statistics_options_layout(
            selected_visualization: VisualizationOptions,
        ) -> List[dict]:
            """Only show statistics checklist if in statistics mode"""

            def get_style(visualization_options: List[VisualizationOptions]) -> dict:
                return (
                    {"display": "block"}
                    if selected_visualization in visualization_options
                    else {"display": "none"}
                )

            statistics_options_style = get_style(
                [
                    VisualizationOptions.STATISTICS,
                    VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                ]
            )
            fanchart_options_style = get_style([VisualizationOptions.FANCHART])

            return [statistics_options_style, fanchart_options_style]



from .dialog_controller import open_dialogs
from .intersection_controller import update_intersection
from .intersection_source_controller import update_intersection_source
from .map_controller import update_maps
from .realization_filter_controller import update_realizations
from .uncertainty_table_controller import update_uncertainty_table



from typing import Dict, List, Union

from dash.dash_table.Format import Format

from ._tornado_data import TornadoData


class TornadoTable:
    """Creates `data`for a `dash_table.DataTable` from a TornadaData instance."""

    def __init__(
        self,
        tornado_data: TornadoData,
        use_si_format: bool = True,
        precision: int = 4,
    ) -> None:
        self._table = tornado_data.tornadotable.copy()
        self.scale = tornado_data.scale
        self._use_si_format = use_si_format
        self._precision = precision
        self._table["low_reals"] = self._table["low_reals"].apply(lambda x: str(len(x)))
        self._table["high_reals"] = self._table["high_reals"].apply(
            lambda x: str(len(x))
        )
        self._table["Response"] = tornado_data.response_name
        self._table.rename(
            columns={
                "sensname": "Sensitivity",
                "low": "Delta low" + (" (%)" if self.scale == "Percentage" else ""),
                "high": "Delta high" + (" (%)" if self.scale == "Percentage" else ""),
                "true_low": "True low",
                "true_high": "True high",
                "low_reals": "Low #reals",
                "high_reals": "High #reals",
            },
            inplace=True,
        )

    def set_column_format(self, col: str) -> Union[dict, Format]:
        if "%" in col:
            return {"specifier": ".1f"}
        if self._use_si_format:
            return {
                "locale": {"symbol": ["", ""]},
                "specifier": f"$.{self._precision}s",
            }
        return Format(precision=self._precision)

    @property
    def columns(self) -> List[Dict]:
        return [
            {
                "name": col,
                "id": col,
                "type": "numeric",
                "format": self.set_column_format(col),
            }
            for col in [
                "Response",
                "Sensitivity",
                "Delta low" + (" (%)" if self.scale == "Percentage" else ""),
                "Delta high" + (" (%)" if self.scale == "Percentage" else ""),
                "True low",
                "True high",
                "Low #reals",
                "High #reals",
            ]
        ]

    @property
    def as_plotly_table(self) -> list:
        return self._table.iloc[::-1].to_dict("records")



from pathlib import Path

import numpy as np
import pytest

from webviz_subsurface._models.ensemble_set_model import EnsembleSetModel


@pytest.mark.usefixtures("app")
def test_single_ensemble(testdata_folder):

    emodel = EnsembleSetModel(
        ensemble_paths={
            "iter-0": str(
                Path(testdata_folder / "01_drogon_ahm" / "realization-*" / "iter-0")
            )
        }
    )
    assert emodel.ens_folders == {"iter-0": Path(testdata_folder) / "01_drogon_ahm"}
    smry = emodel.get_or_load_smry_cached()
    assert len(smry.columns) == 934
    assert len(smry["DATE"].unique()) == 2369
    assert smry["ENSEMBLE"].unique() == ["iter-0"]
    assert smry["ENSEMBLE"].dtype == np.dtype("O")
    assert all(
        np.issubdtype(dtype, np.number)
        for dtype in smry.drop(["REAL", "ENSEMBLE", "DATE"], axis=1).dtypes
    )

    parameters = emodel.load_parameters()
    assert all(col in parameters.columns for col in ["ENSEMBLE", "REAL"])
    assert parameters["ENSEMBLE"].dtype == np.dtype("O")
    assert parameters["REAL"].dtype == np.dtype("int64")
    assert len(parameters.columns) == 56


@pytest.mark.usefixtures("app")
def test_smry_load_multiple_ensembles(testdata_folder):

    emodel = EnsembleSetModel(
        ensemble_paths={
            "iter-0": str(
                Path(testdata_folder / "01_drogon_ahm" / "realization-*" / "iter-0")
            ),
            "iter-3": str(
                Path(testdata_folder / "01_drogon_ahm" / "realization-*" / "iter-3")
            ),
        }
    )
    smry = emodel.get_or_load_smry_cached()
    assert len(smry.columns) == 934
    assert len(smry["DATE"].unique()) == 4167
    assert set(smry["ENSEMBLE"].unique()) == set(["iter-0", "iter-3"])
    assert smry["ENSEMBLE"].dtype == np.dtype("O")
    # assert smry["DATE"].dtype == np.dtype("O") # Fails due to wrong input data?
    assert smry["REAL"].dtype == np.dtype("int64")
    assert all(
        np.issubdtype(dtype, np.number)
        for dtype in smry.drop(["REAL", "ENSEMBLE", "DATE"], axis=1).dtypes
    )
    smeta = emodel.load_smry_meta()
    assert set(smeta.columns) == set(
        ["unit", "is_total", "is_rate", "is_historical", "keyword", "wgname", "get_num"]
    )
    assert len(smeta) == 931
    assert "FOPT" in smeta.index

    parameters = emodel.load_parameters()
    assert all(col in parameters.columns for col in ["ENSEMBLE", "REAL"])
    assert parameters["ENSEMBLE"].dtype == np.dtype("O")
    assert parameters["REAL"].dtype == np.dtype("int64")
    assert len(parameters.columns) == 56

    dframe = emodel.load_csv(Path("share") / "results" / "tables" / "rft.csv")
    assert "ENSEMBLE" in dframe.columns
    assert dframe["ENSEMBLE"].dtype == np.dtype("O")
    assert len(dframe["ENSEMBLE"].unique()) == 2
    assert len(dframe.columns) == 15

    with pytest.raises(KeyError) as exc:
        emodel.load_csv("some_path")
    assert (
        exc.value.args[0]
        == "No data found for load_csv with arguments: {'csv_file': 'some_path'}"
    )


@pytest.mark.usefixtures("app")
def test_webvizstore(testdata_folder):
    emodel = EnsembleSetModel(
        ensemble_paths={
            "iter-0": str(
                Path(testdata_folder / "01_drogon_ahm" / "realization-*" / "iter-0")
            ),
            "iter-3": str(
                Path(testdata_folder / "01_drogon_ahm" / "realization-*" / "iter-3")
            ),
        }
    )
    emodel.load_parameters()
    assert len(emodel.webvizstore) == 2
    emodel.get_or_load_smry_cached()
    assert len(emodel.webvizstore) == 4
    emodel.load_smry_meta()
    assert len(emodel.webvizstore) == 6
    emodel.load_csv(Path("share") / "results" / "tables" / "rft.csv")
    assert len(emodel.webvizstore) == 8



import colorsys
from typing import Tuple

from webviz_config.utils import StrEnum


class StandardColors(StrEnum):
    OIL_GREEN = "#2ca02c"
    WATER_BLUE = "#1f77b4"
    GAS_RED = "#d62728"


def hex_to_rgb(hex_string: str) -> Tuple[float, float, float]:
    """Converts the given hex color to rgb tuple with floating point byte color values.

    Byte color channels: 0-255

    `Return:`
    RGB color on tuple format Tuple[float, float, float] with r-, g- and b-channel
    on index 0, 1 and 2, respectively. With floating point byte color value 0-255.
    """
    hex_string = hex_string.lstrip("#")
    hlen = len(hex_string)
    rgb = [int(hex_string[i : i + hlen // 3], 16) for i in range(0, hlen, hlen // 3)]
    return float(rgb[0]), float(rgb[1]), float(rgb[2])


def hex_to_rgba(
    hex_string: str, opacity: float = 1.0
) -> Tuple[float, float, float, float]:
    """Converts the given hex color to rgba tuple with floating point byte color values
    and alpha channel as opacity.

    Byte color channels: 0-255
    alpha: 0-1

    `Return:`
    RGBA color on tuple format Tuple[float,float,float,float] with r-, g-, b- and alpha-channel
    on index 0, 1, 2 and 3, respectively. With floating point byte color value 0-255.
    """
    rgb = hex_to_rgb(hex_string)
    alpha = max(0.0, min(1.0, opacity))
    return rgb[0], rgb[1], rgb[2], alpha


def hex_to_rgb_str(hex_string: str) -> str:
    """Converts the given hex color to rgb string

    Byte color channels: 0-255

    `Return:`
    RGB color on string format "rgb(r,g,b)" where, channels r, g and b are
    represented with byte color value 0-255.
    """
    rgb = hex_to_rgb(hex_string)
    return f"rgb({round(rgb[0])}, {round(rgb[1])}, {round(rgb[2])})"


def hex_to_rgba_str(hex_string: str, opacity: float = 1.0) -> str:
    """Converts the given hex color to rgba string

    Byte color channels: 0-255

    `Return:`
    RGB color on string format "rgba(r,g,b,alpha)" where, channels r, g and b are
    represented with byte color value 0-255.
    """
    rgba = hex_to_rgba(hex_string, opacity)
    return f"rgba({rgba[0]}, {rgba[1]}, {rgba[2]}, {rgba[3]})"


def rgb_to_str(rgb: Tuple[float, float, float]) -> str:
    """Convert rgb tuple with floating point byte color values to string

    `Input:`
    * rgb - Tuple[float,float,float] - RGB color on tuple format, r-, g- and b-channel
    is index 0, 1 and 2, respectively. Byte color values 0-255

    `Return:`
    RGB color on string format "rgb(r,g,b)" where, channels r, g and b are
    represented with byte color integer value 0-255.
    """
    return f"rgb({round(rgb[0])}, {round(rgb[1])}, {round(rgb[2])})"


def rgba_to_str(rgba: Tuple[float, float, float, float]) -> str:
    """Convert rgba tuple with floating point byte color values to string

    `Input:`
    * rgb - Tuple[float,float,float,float] - RGBA color on tuple format

    `Return:`
    RGBA color on string format "rgba(r,g,b,a)" where, channels r, g and b are
    represented with byte color integer value 0-255, and a is represented by a
    decimal number 0-1"""
    return f"rgba({round(rgba[0])}, {round(rgba[1])}, {round(rgba[2])}, {rgba[3]})"


def rgba_to_tuple(rgba: str) -> Tuple[float, float, float, float]:
    """
    Takes rgba color 'rgba(a, b, c, d)' and returns a tuple (a, b, c, d)
    """
    numbers = [float(value) for value in rgba.strip("rbga()").split(",")]
    return numbers[0], numbers[1], numbers[2], numbers[3]


def rgba_to_hex(color: str) -> str:
    """Converts a rgba color to hex"""
    color_list = color.strip("rgba()").split(",")
    return "#" + "".join(f"{int(i):02x}" for i in color_list)


def scale_rgb_lightness(
    rgb: Tuple[float, float, float],
    scale_percentage: float,
    min_lightness_percentage: float = 10.0,
    max_lightness_percentage: float = 90.0,
) -> Tuple[float, float, float]:
    """Scale lightness of rgb tuple with byte color values, in percentage

    Method utilizes HLS color space, and adjust lightness of color
    where larger percentage is lighther, and lower percentage is darker.

    `Input:`
    * rgb - Tuple[float,float,float] - RGB color on tuple format, r-, g- and b-channel
    is index 0, 1 and 2, respectively. Byte color values 0-255
    * scale_percentage: float - Color scaling in percentage
    * min_lightness_percentage: float - Minimum percent lightness, to prevent black color
    * max_lightness_percentage: float - Maximum percent lightness, to prevent white color
    """
    # Convert scale to scalar
    l_scale = scale_percentage / 100.0

    # Convert min and max lightness from percentage to scalar 0-1
    l_min = max(0.0, min_lightness_percentage / 100.0)
    l_max = min(1.0, max_lightness_percentage / 100.0)

    # Convert rgb to hls
    hls = colorsys.rgb_to_hls(
        float(rgb[0]) / 255.0, float(rgb[1]) / 255.0, float(rgb[2]) / 255.0
    )

    # Scale lightness within min and max
    l_scaled = min(l_max, max(l_min, hls[1] * l_scale))

    # Convert lightness scaled hls to rgb
    rgb = colorsys.hls_to_rgb(hls[0], l_scaled, s=hls[2])
    return 255.0 * rgb[0], 255.0 * rgb[1], 255.0 * rgb[2]


def find_intermediate_color(lowcolor: str, highcolor: str, intermed: float) -> str:
    """
    Returns the color at a given distance between two colors
    This function takes two color tuples, where each element is between 0
    and 1, along with a value 0 < intermed < 1 and returns a color that is
    intermed-percent from lowcolor to highcolor.
    """

    # convert to tuple color, eg. (1, 0.45, 0.7)
    lowcolor_tuple = rgba_to_tuple(lowcolor)
    highcolor_tuple = rgba_to_tuple(highcolor)

    diff_0 = float(highcolor_tuple[0] - lowcolor_tuple[0])
    diff_1 = float(highcolor_tuple[1] - lowcolor_tuple[1])
    diff_2 = float(highcolor_tuple[2] - lowcolor_tuple[2])
    diff_3 = float(highcolor_tuple[3] - lowcolor_tuple[3])

    inter_med_tuple = (
        lowcolor_tuple[0] + intermed * diff_0,
        lowcolor_tuple[1] + intermed * diff_1,
        lowcolor_tuple[2] + intermed * diff_2,
        lowcolor_tuple[3] + intermed * diff_3,
    )

    # back to an rgba string, e.g. rgba(30, 20, 10)
    return rgba_to_str(inter_med_tuple)



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

import pandas as pd
import webviz_core_components as wcc
from dash import (
    ClientsideFunction,
    Dash,
    Input,
    Output,
    State,
    callback_context,
    dash_table,
    dcc,
    html,
)
from dash.exceptions import PreventUpdate
from webviz_config import WebvizSettings
from webviz_config.webviz_assets import WEBVIZ_ASSETS

import webviz_subsurface

from ._tornado_bar_chart import TornadoBarChart
from ._tornado_data import TornadoData
from ._tornado_table import TornadoTable


class TornadoWidget:
    """### TornadoWidget

    This component visualizes a Tornado plot.
    It is meant to be used as a component in other plugin, and is initialized
     with a dataframe of realizations with corresponding sensitivities,
    but without the response values that are to be plotted.
    Instead we registers a dcc.Store which will contain the response values.

    To use:
    1. Initialize an instance of this class in a plugin.
    2. Add tornadoplot.layout to the plugin layout
    3. Register a callback that writes a json dump to tornadoplot.storage_id
    The format of the json dump must be ('ENSEMBLE' and 'data' are mandatory, the others optional):
    {'ENSEMBLE': name of ensemble,
     'data': 2d array of realizations / response values
     'number_format' (str): Format of the numeric part based on the Python Format Specification
      Mini-Language e.g. '#.3g' for 3 significant digits, '.2f' for two decimals, or '.0f' for no
      decimals.
     'unit' (str): String to append at the end as a unit.
     'spaced' (bool): Include a space between last numerical digit and SI-prefix.
     'locked_si_prefix' (str or int): Lock the SI prefix to either a string (e.g. 'm' (milli) or 'M'
      (mega)), or an integer which is the base 10 exponent (e.g. 3 for kilo, -3 for milli).
    }

    Mouse events:
    The current case at mouse cursor can be retrieved by registering a callback
    that reads from  `tornadoplot.click_id` if `allow_click` has been specified at initialization.


    * `realizations`: Dataframe of realizations with corresponding sensitivity cases
    * `reference`: Which sensitivity to use as reference.
    * `allow_click`: Registers a callback to store current data on mouse click
    """

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        realizations: pd.DataFrame,
        reference: str = "rms_seed",
        allow_click: bool = False,
    ):
        self.realizations = realizations
        self.sensnames = list(self.realizations["SENSNAME"].unique())
        if self.sensnames == [None]:
            raise KeyError(
                "No sensitivity information found in ensemble. "
                "Containers utilizing tornadoplot can only be used for ensembles with "
                "one by one design matrix setup "
                "(SENSNAME and SENSCASE must be present in parameter file)."
            )
        self.initial_reference = (
            reference if reference in self.sensnames else self.sensnames[0]
        )
        self.allow_click = allow_click
        self.uid = uuid4()
        self.plotly_theme = webviz_settings.theme.plotly_theme
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent
            / "_assets"
            / "js"
            / "clientside_functions.js"
        )
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self) -> List[Dict[str, Any]]:
        return [
            {
                "id": self.ids("tornado-graph"),
                "content": ("Shows tornado plot."),
            },
            {
                "id": self.ids("reference"),
                "content": (
                    "Set reference sensitivity for which to calculate tornado plot"
                ),
            },
            {
                "id": self.ids("scale"),
                "content": (
                    "Set tornadoplot scale to either percentage or absolute values"
                ),
            },
            {
                "id": self.ids("cut-by-ref"),
                "content": (
                    "Remove sensitivities smaller than the reference from the plot"
                ),
            },
            {
                "id": self.ids("reset"),
                "content": "Clears the currently selected sensitivity",
            },
        ]

    @property
    def storage_id(self) -> str:
        """The id of the dcc.Store component that holds the tornado data"""
        return self.ids("storage")

    @property
    def click_id(self) -> str:
        """The id of the dcc.Store component that holds click data"""
        return self.ids("click-store")

    @property
    def high_low_storage_id(self) -> str:
        """The id of the dcc.Store component that holds click data"""
        return self.ids("high-low-storage")

    @property
    def settings_layout(self) -> html.Div:
        return html.Div(
            children=[
                html.Div(
                    children=[
                        wcc.FlexBox(
                            style={"fontSize": "15px"},
                            children=[
                                html.Div(
                                    style={
                                        "minWidth": "100px",
                                        "flex": 2,
                                        "marginRight": "10px",
                                    },
                                    children=[
                                        wcc.Dropdown(
                                            label="Reference",
                                            id=self.ids("reference"),
                                            options=[
                                                {
                                                    "label": r,
                                                    "value": r,
                                                }
                                                for r in self.sensnames
                                            ],
                                            value=self.initial_reference,
                                            clearable=False,
                                        ),
                                        wcc.Dropdown(
                                            label="Scale",
                                            id=self.ids("scale"),
                                            options=[
                                                {
                                                    "label": r,
                                                    "value": r,
                                                }
                                                for r in [
                                                    "Relative value (%)",
                                                    "Relative value",
                                                    "True value",
                                                ]
                                            ],
                                            value="Relative value (%)",
                                            clearable=False,
                                        ),
                                        html.Button(
                                            style={
                                                "fontSize": "10px",
                                                "marginTop": "10px",
                                            }
                                            if self.allow_click
                                            else {"display": "none"},
                                            id=self.ids("reset"),
                                            children="Clear selected",
                                        ),
                                    ],
                                ),
                                html.Div(
                                    style={
                                        "minWidth": "100px",
                                        "flex": 2,
                                        "marginRight": "10px",
                                    },
                                    children=[
                                        wcc.SelectWithLabel(
                                            label="Select sensitivities",
                                            id=self.ids("sens_filter"),
                                            options=[
                                                {
                                                    "label": i,
                                                    "value": i,
                                                }
                                                for i in self.sensnames
                                            ],
                                            value=self.sensnames,
                                            multi=True,
                                            size=min(
                                                8,
                                                len(self.sensnames),
                                            ),
                                        ),
                                    ],
                                ),
                                html.Div(
                                    style={
                                        "minWidth": "100px",
                                        "flex": 3,
                                    },
                                    children=[
                                        wcc.Checklist(
                                            label="Plot options",
                                            id=self.ids("plot-options"),
                                            options=[
                                                {
                                                    "label": "Fit all bars in figure",
                                                    "value": "Fit all bars in figure",
                                                },
                                                {
                                                    "label": "Remove sensitivites with no impact",
                                                    "value": "Remove sensitivites with no impact",
                                                },
                                                {
                                                    "label": "Show realization points",
                                                    "value": "Show realization points",
                                                },
                                                {
                                                    "label": "Color bars by sensitivity",
                                                    "value": "Color bars by sensitivity",
                                                },
                                            ],
                                            value=[],
                                            labelStyle={"display": "block"},
                                        ),
                                        wcc.Dropdown(
                                            label="Label",
                                            id=self.ids("label"),
                                            options=[
                                                {"label": "No label", "value": "hide"},
                                                {
                                                    "label": "Simple label",
                                                    "value": "simple",
                                                },
                                                {
                                                    "label": "Detailed label",
                                                    "value": "detailed",
                                                },
                                            ],
                                            value="detailed",
                                            clearable=False,
                                        ),
                                    ],
                                ),
                            ],
                        ),
                    ]
                ),
            ],
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            style={"marginLeft": "10px", "height": "90vh"},
            children=[
                html.Div(
                    children=[
                        html.Label(
                            "Tornado Plot",
                            style={
                                "textAlign": "center",
                                "font-weight": "bold",
                            },
                        ),
                        wcc.RadioItems(
                            vertical=False,
                            id=self.ids("plot-or-table"),
                            options=[
                                {"label": "Show bars", "value": "bars"},
                                {
                                    "label": "Show table",
                                    "value": "table",
                                },
                            ],
                            value="bars",
                        ),
                    ],
                ),
                html.Div(
                    style={"overflowY": "auto", "height": "60vh"},
                    children=[
                        html.Div(
                            id=self.ids("graph-wrapper"),
                            style={},
                            children=wcc.Graph(
                                id=self.ids("tornado-graph"),
                                config={"displayModeBar": False},
                            ),
                        ),
                        html.Div(
                            id=self.ids("table-wrapper"),
                            style={"display": "none"},
                            children=dash_table.DataTable(
                                id=self.ids("tornado-table"),
                                style_cell={
                                    "whiteSpace": "normal",
                                    "height": "auto",
                                },
                            ),
                        ),
                    ],
                ),
                html.Hr(style={"marginTop": "10px"}),
                self.settings_layout,
                dcc.Store(id=self.ids("storage"), storage_type="session"),
                dcc.Store(id=self.ids("click-store"), storage_type="session"),
                dcc.Store(id=self.ids("high-low-storage"), storage_type="session"),
                dcc.Store(id=self.ids("client-height-pixels"), storage_type="session"),
            ],
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.ids("label"), "disabled"),
            Input(self.ids("plot-options"), "value"),
        )
        def _disable_label(plot_options: List) -> bool:
            return "Show realization points" in plot_options

        @app.callback(
            Output(self.ids("graph-wrapper"), "style"),
            Output(self.ids("table-wrapper"), "style"),
            Input(self.ids("plot-or-table"), "value"),
            State(self.ids("graph-wrapper"), "style"),
            State(self.ids("table-wrapper"), "style"),
        )
        def _set_visualization(
            viz_type: str, graph_style: dict, table_style: dict
        ) -> Tuple[Dict[str, str], Dict[str, str]]:
            if viz_type == "bars":
                graph_style.update({"display": "inline"})
                table_style.update({"display": "none"})
            if viz_type == "table":
                graph_style.update({"display": "none"})
                table_style.update({"display": "inline"})
            return graph_style, table_style

        app.clientside_callback(
            ClientsideFunction(
                namespace="clientside", function_name="get_client_height"
            ),
            Output(self.ids("client-height-pixels"), "data"),
            Input(self.ids("plot-options"), "value"),
        )

        @app.callback(
            Output(self.ids("tornado-graph"), "figure"),
            Output(self.ids("tornado-table"), "data"),
            Output(self.ids("tornado-table"), "columns"),
            Output(self.ids("high-low-storage"), "data"),
            Input(self.ids("reference"), "value"),
            Input(self.ids("scale"), "value"),
            Input(self.ids("plot-options"), "value"),
            Input(self.ids("label"), "value"),
            Input(self.ids("storage"), "data"),
            Input(self.ids("sens_filter"), "value"),
            State(self.ids("client-height-pixels"), "data"),
        )
        def _calc_tornado(
            reference: str,
            scale: str,
            plot_options: List,
            label_option: str,
            data: Union[str, bytes, bytearray],
            sens_filter: List[str],
            client_height: Optional[int],
        ) -> Tuple[Any, List[Any], List[Dict[Any, Any]], Dict[str, Dict[Any, Any]]]:
            if not data:
                raise PreventUpdate
            plot_options = plot_options if plot_options else []
            data = json.loads(data)
            if not isinstance(data, dict):
                raise PreventUpdate
            values = pd.DataFrame(data["data"], columns=["REAL", "VALUE"])
            realizations = self.realizations.loc[
                self.realizations["ENSEMBLE"] == data["ENSEMBLE"]
            ]

            design_and_responses = pd.merge(values, realizations, on="REAL")
            if sens_filter is not None:
                if reference not in sens_filter:
                    sens_filter.append(reference)
                design_and_responses = design_and_responses.loc[
                    design_and_responses["SENSNAME"].isin(sens_filter)
                ]
            tornado_data = TornadoData(
                dframe=design_and_responses,
                response_name=data.get("response_name"),
                reference=reference,
                scale="Percentage" if scale == "Relative value (%)" else "Absolute",
                cutbyref="Remove sensitivites with no impact" in plot_options,
            )

            figure_height = (
                client_height * 0.59
                if "Fit all bars in figure" in plot_options
                and client_height is not None
                else max(100 * len(tornado_data.tornadotable["sensname"].unique()), 200)
            )
            tornado_figure = TornadoBarChart(
                tornado_data=tornado_data,
                plotly_theme=self.plotly_theme,
                figure_height=figure_height,
                label_options=label_option,
                number_format=data.get("number_format", ""),
                unit=data.get("unit", ""),
                spaced=data.get("spaced", True),
                locked_si_prefix=data.get("locked_si_prefix", None),
                use_true_base=scale == "True value",
                show_realization_points="Show realization points" in plot_options,
                color_by_sensitivity="Color bars by sensitivity" in plot_options,
            )
            tornado_table = TornadoTable(tornado_data=tornado_data)
            return (
                tornado_figure.figure,
                tornado_table.as_plotly_table,
                tornado_table.columns,
                tornado_data.low_high_realizations_list,
            )

        if self.allow_click:

            @app.callback(
                Output(self.ids("click-store"), "data"),
                [
                    Input(self.ids("tornado-graph"), "clickData"),
                    Input(self.ids("reset"), "n_clicks"),
                    State(self.ids("high-low-storage"), "data"),
                ],
            )
            def _save_click_data(
                data: dict, nclicks: Optional[int], sens_reals: dict
            ) -> str:
                if (
                    callback_context.triggered is None
                    or sens_reals is None
                    or data is None
                ):
                    raise PreventUpdate
                ctx = callback_context.triggered[0]["prop_id"].split(".")[0]
                if ctx == self.ids("reset") and nclicks:
                    return json.dumps(
                        {
                            "real_low": [],
                            "real_high": [],
                            "sens_name": None,
                        }
                    )
                sensname = data["points"][0]["y"]
                real_high = sens_reals[sensname]["real_high"]
                real_low = sens_reals[sensname]["real_low"]
                return json.dumps(
                    {
                        "real_low": real_low,
                        "real_high": real_high,
                        "sens_name": sensname,
                    }
                )



from typing import Dict, List, Optional, Tuple

import webviz_core_components as wcc
from dash import Input, Output, State, callback
from dash.development.base_component import Component
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import DepthType, LineType
from ...._utils import RftPlotterDataModel, filter_frame


class FormationPlotSettings(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLE = "ensemble"
        WELL = "well"
        DATE = "date"
        LINETYPE = "linetype"
        DEPTH_OPTION = "depth-option"

    def __init__(self, datamodel: RftPlotterDataModel) -> None:
        super().__init__("Formation plot settings")
        self._datamodel = datamodel
        self._ensembles = datamodel.ensembles
        self._well_names = datamodel.well_names

    def layout(self) -> List[Component]:
        ensemble = self._ensembles[0] if self._ensembles else None
        well = self._well_names[0] if self._well_names else None
        dates_in_well = self._datamodel.date_in_well(well) if well is not None else []
        return [
            wcc.Dropdown(
                label="Ensemble",
                id=self.register_component_unique_id(self.Ids.ENSEMBLE),
                options=[{"label": ens, "value": ens} for ens in self._ensembles],
                value=[ensemble],
                multi=True,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Well",
                id=self.register_component_unique_id(self.Ids.WELL),
                options=[{"label": well, "value": well} for well in self._well_names],
                value=well,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Date",
                id=self.register_component_unique_id(self.Ids.DATE),
                options=[{"label": date, "value": date} for date in dates_in_well],
                clearable=False,
                value=dates_in_well[0] if dates_in_well else None,
            ),
            wcc.RadioItems(
                label="Plot simulations as",
                id=self.register_component_unique_id(self.Ids.LINETYPE),
                options=[
                    {
                        "label": "Realization lines",
                        "value": LineType.REALIZATION,
                    },
                    {
                        "label": "Statistical fanchart",
                        "value": LineType.FANCHART,
                    },
                ],
                value=LineType.REALIZATION,
            ),
            wcc.RadioItems(
                label="Depth option",
                id=self.register_component_unique_id(self.Ids.DEPTH_OPTION),
                options=[
                    {
                        "label": "TVD",
                        "value": DepthType.TVD,
                    },
                    {
                        "label": "MD",
                        "value": DepthType.MD,
                    },
                ],
                value=DepthType.TVD,
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(self.Ids.LINETYPE).to_string(),
                "options",
            ),
            Output(
                self.component_unique_id(self.Ids.LINETYPE).to_string(),
                "value",
            ),
            Input(
                self.component_unique_id(self.Ids.DEPTH_OPTION).to_string(),
                "value",
            ),
            State(
                self.component_unique_id(self.Ids.LINETYPE).to_string(),
                "value",
            ),
            State(self.component_unique_id(self.Ids.WELL).to_string(), "value"),
            State(self.component_unique_id(self.Ids.DATE).to_string(), "value"),
        )
        @callback_typecheck
        def _update_linetype(
            depthtype: DepthType,
            current_linetype: LineType,
            current_well: str,
            current_date: str,
        ) -> Tuple[List[Dict[str, str]], str]:
            if self._datamodel.simdf is not None:
                df = filter_frame(
                    self._datamodel.simdf,
                    {"WELL": current_well, "DATE": current_date},
                )
                if depthtype == DepthType.TVD or (
                    depthtype == DepthType.MD
                    and "CONMD" in self._datamodel.simdf
                    and len(df["CONMD"].unique()) == len(df["DEPTH"].unique())
                ):

                    return [
                        {
                            "label": "Realization lines",
                            "value": LineType.REALIZATION,
                        },
                        {
                            "label": "Statistical fanchart",
                            "value": LineType.FANCHART,
                        },
                    ], current_linetype

            return [
                {
                    "label": "Realization lines",
                    "value": LineType.REALIZATION,
                },
            ], LineType.REALIZATION

        @callback(
            Output(
                self.component_unique_id(self.Ids.DATE).to_string(),
                "options",
            ),
            Output(self.component_unique_id(self.Ids.DATE).to_string(), "value"),
            Input(self.component_unique_id(self.Ids.WELL).to_string(), "value"),
            State(self.component_unique_id(self.Ids.DATE).to_string(), "value"),
        )
        @callback_typecheck
        def _update_date(
            well: str, current_date: str
        ) -> Tuple[List[Dict[str, str]], Optional[str]]:
            dates = self._datamodel.date_in_well(well)
            first_date = dates[0] if dates else None
            available_dates = [{"label": date, "value": date} for date in dates]
            date = current_date if current_date in dates else first_date
            return available_dates, date



# pylint: disable=protected-access
import pandas as pd
import pytest

from webviz_subsurface._components.tornado._tornado_data import TornadoData


def test_tornado_data_init():
    # fmt: off
    input_data = [
        ["REAL",  "SENSNAME",   "SENSCASE",  "SENSTYPE", "VALUE" ],
        [     0,         "A",    "p10_p90", "mc"       ,     10.0],
        [     1,         "A",    "p10_p90", "mc"       ,     20.0],
        [     2,         "B",       "deep", "scalar"   ,     5.0],
        [     3,         "B",       "deep", "scalar"   ,     6.0],
        [     4,         "C",    "shallow", "scalar"   ,     25.0],
        [     5,         "C",    "shallow", "scalar"   ,     26.0],
        [     6,         "D", "simulation", "mc"       ,     9.0],
        [     7,         "D", "simulation", "mc"       ,     11.0],
    ]
    # fmt: on

    input_df = pd.DataFrame(input_data[1:], columns=input_data[0])

    with pytest.raises(ValueError) as exc:
        TornadoData(dframe=input_df)
    assert exc.value.args[0] == "Reference SENSNAME rms_seed not in input data"

    tornado_data = TornadoData(dframe=input_df, reference="A")

    assert tornado_data._calculate_ref_average(input_df) == 15.0
    avg_list = tornado_data._calculate_sensitivity_averages(input_df)
    assert avg_list[0] == {
        "sensname": "A",
        "senscase": "P90",
        "values": 11.0,
        "values_ref": -26.666666666666668,
        "reals": [0],
    }
    low_high_list = tornado_data._calculate_tornado_low_high_list(avg_list)
    assert low_high_list[0] == {
        "low": -26.666666666666668,
        "low_base": 0,
        "low_label": "P90",
        "low_tooltip": -26.666666666666668,
        "true_low": 11.0,
        "low_reals": [0],
        "sensname": "A",
        "high": 26.666666666666668,
        "high_base": 0,
        "high_label": "P10",
        "high_tooltip": 26.666666666666668,
        "true_high": 19.0,
        "high_reals": [1],
    }

    tornado_data = TornadoData(
        dframe=input_df, reference="B", scale="Absolute", cutbyref=True
    )
    assert tornado_data._calculate_tornado_table(input_df).to_dict("records") == [
        {
            "low": 0.0,
            "low_base": 5.5,
            "low_label": "P90",
            "low_tooltip": 5.5,
            "true_low": 11.0,
            "low_reals": [],
            "sensname": "A",
            "high": 8.0,
            "high_base": 5.5,
            "high_label": "P10",
            "high_tooltip": 13.5,
            "true_high": 19.0,
            "high_reals": [0, 1],
        },
        {
            "low": 0.0,
            "low_base": 0.0,
            "low_label": None,
            "low_tooltip": 0.0,
            "true_low": 5.5,
            "low_reals": [],
            "sensname": "B",
            "high": 0.0,
            "high_base": 0.0,
            "high_label": "deep",
            "high_tooltip": 0.0,
            "true_high": 5.5,
            "high_reals": [2, 3],
        },
        {
            "low": 0.0,
            "low_base": 0.0,
            "low_label": None,
            "low_tooltip": 0.0,
            "true_low": 5.5,
            "low_reals": [],
            "sensname": "C",
            "high": 20.0,
            "high_base": 0.0,
            "high_label": "shallow",
            "high_tooltip": 20.0,
            "true_high": 25.5,
            "high_reals": [4, 5],
        },
        {
            "low": 0.0,
            "low_base": 3.6999999999999993,
            "low_label": "P90",
            "low_tooltip": 3.6999999999999993,
            "true_low": 9.2,
            "low_reals": [],
            "sensname": "D",
            "high": 1.6000000000000014,
            "high_base": 3.6999999999999993,
            "high_label": "P10",
            "high_tooltip": 5.300000000000001,
            "true_high": 10.8,
            "high_reals": [6, 7],
        },
    ]



from typing import List

import dash_bootstrap_components as dbc
import webviz_core_components as wcc
from dash import html


def dialog_layout(
    uuid: str,
    dialog_id: str,
    title: str,
    children: List,
    size: str = "sm",
) -> wcc.Dialog:

    return wcc.Dialog(
        # style={"marginTop": "20vh"},
        children=children,
        id={"id": uuid, "dialog_id": dialog_id, "element": "wrapper"},
        max_width=size,
        title=title,
        draggable=True,
    )


def open_dialog_layout(uuid: str, dialog_id: str, title: str) -> dbc.Button:
    return html.Div(
        children=html.Button(
            title,
            id={"id": uuid, "dialog_id": dialog_id, "element": "button-open"},
        ),
    )


def clear_all_apply_dialog_buttons(
    uuid: str, dialog_id: str, apply_disabled: bool = True
) -> html.Div:
    return html.Div(
        children=[
            dbc.Button(
                "Clear",
                style={"padding": "0 20px"},
                className="mr-1",
                id={"id": uuid, "dialog_id": dialog_id, "element": "clear"},
            ),
            dbc.Button(
                "All",
                style={"padding": "0 20px"},
                className="mr-1",
                id={"id": uuid, "dialog_id": dialog_id, "element": "all"},
            ),
            dbc.Button(
                "Apply",
                style={"padding": "0 20px", "visibility": "hidden"}
                if apply_disabled
                else {"padding": "0 20px"},
                className="mr-1",
                id={"id": uuid, "dialog_id": dialog_id, "element": "apply"},
            ),
        ]
    )



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class ParameterSettings(SettingsGroupABC):
    class IDs:
        # pylint: disable=too-few-public-methods
        SHARED_ENSEMBLE = "shared-ensemble"
        PARAMETER_H = "parameter-horizontal"
        ENSEMBLE_H = "ensemble-horizontal"
        PARAMETER_V = "parameter-vertical"
        ENSEMBLE_V = "ensemble-vertical"
        SCATTER_VISIBLE = "scatter-visible"
        SCATTER_COLOR = "scatter-color"

    def __init__(self, ensembles: dict, p_cols: List) -> None:
        super().__init__("Settings")
        self.ensembles = ensembles
        self.p_cols = p_cols

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(
                    ParameterSettings.IDs.SHARED_ENSEMBLE
                ),
                label="Ensemble in both plot",
                options=[{"label": k, "value": v} for k, v in self.ensembles.items()],
                value=list(self.ensembles.values())[0],
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(ParameterSettings.IDs.PARAMETER_H),
                label="Parameter (horizontal axis)",
                options=[{"label": p, "value": p} for p in self.p_cols],
                value=self.p_cols[0] if len(self.p_cols) > 0 else "",
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(ParameterSettings.IDs.ENSEMBLE_H),
                label="Ensemble (horizontal axis)",
                options=[{"label": k, "value": v} for k, v in self.ensembles.items()],
                value=list(self.ensembles.values())[0]
                if len(self.ensembles.values()) > 0
                else "",
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(ParameterSettings.IDs.PARAMETER_V),
                label="Parameter (vertical axis)",
                options=[{"label": p, "value": p} for p in self.p_cols],
                value=self.p_cols[0] if len(self.p_cols) > 0 else "",
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(ParameterSettings.IDs.ENSEMBLE_V),
                label="Ensemble (vertical axis)",
                options=[{"label": k, "value": v} for k, v in self.ensembles.items()],
                value=list(self.ensembles.values())[0]
                if len(self.ensembles.values()) > 0
                else "",
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(
                    ParameterSettings.IDs.SCATTER_COLOR
                ),
                label="Color points by",
                options=[{"label": p, "value": p} for p in self.p_cols],
                value=self.p_cols[0] if len(self.p_cols) > 0 else "",
                multi=False,
                clearable=True,
            ),
            wcc.Checklist(
                id=self.register_component_unique_id(
                    ParameterSettings.IDs.SCATTER_VISIBLE
                ),
                style={"padding": "5px"},
                options=[
                    {
                        "label": "Show scatterplot density",
                        "value": "density",
                    }
                ],
            ),
        ]



from pathlib import Path

import pandas as pd
from webviz_config.themes import default_theme

from webviz_subsurface.plugins._property_statistics.models.property_statistics_model import (
    PropertyStatisticsModel,
)


def get_data_df(testdata_folder: Path) -> pd.DataFrame:

    return pd.read_csv(
        testdata_folder
        / "reek_test_data"
        / "aggregated_data"
        / "property_statistics.csv"
    )


def test_init(testdata_folder: Path) -> None:
    data_df = get_data_df(testdata_folder)
    model = PropertyStatisticsModel(dataframe=data_df, theme=default_theme)
    assert set(model.dataframe.columns) == set(
        [
            "PROPERTY",
            "ZONE",
            "REGION",
            "FACIES",
            "Avg",
            "Avg_Weighted",
            "Max",
            "Min",
            "P10",
            "P90",
            "Stddev",
            "SOURCE",
            "ID",
            "REAL",
            "ENSEMBLE",
            "label",
        ]
    )
